from inspect import getinnerframes
from typing import TYPE_CHECKING, Any, Callable, Generator, Optional

import pytest
from _pytest.capture import CaptureFixture
from pytest_mock import MockerFixture
from starlette.exceptions import HTTPException as StarletteHTTPException
from structlog.testing import capture_logs

from litestar import Litestar, MediaType, Request, Response, get
from litestar.exceptions import HTTPException, InternalServerException, ValidationException
from litestar.logging.config import LoggingConfig, StructLoggingConfig
from litestar.middleware.exceptions import ExceptionHandlerMiddleware
from litestar.middleware.exceptions._debug_response import get_symbol_name
from litestar.middleware.exceptions.middleware import _starlette_exception_handler, get_exception_handler
from litestar.status_codes import HTTP_400_BAD_REQUEST, HTTP_500_INTERNAL_SERVER_ERROR
from litestar.testing import TestClient, create_test_client
from litestar.types import ExceptionHandlersMap
from litestar.types.asgi_types import HTTPScope
from tests.helpers import cleanup_logging_impl

if TYPE_CHECKING:
    from _pytest.logging import LogCaptureFixture

    from litestar.types import Scope
    from litestar.types.callable_types import GetLogger


async def dummy_app(scope: Any, receive: Any, send: Any) -> None:
    return None


@pytest.fixture(autouse=True)
def cleanup_logging() -> Generator:
    with cleanup_logging_impl():
        yield


@pytest.fixture()
def app() -> Litestar:
    return Litestar()


@pytest.fixture()
def middleware() -> ExceptionHandlerMiddleware:
    return ExceptionHandlerMiddleware(dummy_app, None, {})


@pytest.fixture()
def scope(create_scope: Callable[..., HTTPScope], app: Litestar) -> HTTPScope:
    return create_scope(app=app)


def test_default_handle_http_exception_handling_extra_object(
    scope: HTTPScope, middleware: ExceptionHandlerMiddleware
) -> None:
    response = middleware.default_http_exception_handler(
        Request(scope=scope), HTTPException(detail="litestar_exception", extra={"key": "value"})
    )
    assert response.status_code == HTTP_500_INTERNAL_SERVER_ERROR
    assert response.content == {
        "detail": "Internal Server Error",
        "extra": {"key": "value"},
        "status_code": 500,
    }


def test_default_handle_http_exception_handling_extra_none(
    scope: HTTPScope, middleware: ExceptionHandlerMiddleware
) -> None:
    response = middleware.default_http_exception_handler(
        Request(scope=scope), HTTPException(detail="litestar_exception")
    )
    assert response.status_code == HTTP_500_INTERNAL_SERVER_ERROR
    assert response.content == {"detail": "Internal Server Error", "status_code": 500}


def test_default_handle_litestar_http_exception_handling(
    scope: HTTPScope, middleware: ExceptionHandlerMiddleware
) -> None:
    response = middleware.default_http_exception_handler(
        Request(scope=scope), HTTPException(detail="litestar_exception")
    )
    assert response.status_code == HTTP_500_INTERNAL_SERVER_ERROR
    assert response.content == {"detail": "Internal Server Error", "status_code": 500}


def test_default_handle_litestar_http_exception_extra_list(
    scope: HTTPScope, middleware: ExceptionHandlerMiddleware
) -> None:
    response = middleware.default_http_exception_handler(
        Request(scope=scope), HTTPException(detail="litestar_exception", extra=["extra-1", "extra-2"])
    )
    assert response.status_code == HTTP_500_INTERNAL_SERVER_ERROR
    assert response.content == {
        "detail": "Internal Server Error",
        "extra": ["extra-1", "extra-2"],
        "status_code": 500,
    }


def test_default_handle_starlette_http_exception_handling(
    scope: HTTPScope, middleware: ExceptionHandlerMiddleware
) -> None:
    response = middleware.default_http_exception_handler(
        Request(scope=scope),
        StarletteHTTPException(detail="litestar_exception", status_code=HTTP_500_INTERNAL_SERVER_ERROR),
    )
    assert response.status_code == HTTP_500_INTERNAL_SERVER_ERROR
    assert response.content == {"detail": "Internal Server Error", "status_code": 500}


def test_default_handle_python_http_exception_handling(
    scope: HTTPScope, middleware: ExceptionHandlerMiddleware
) -> None:
    response = middleware.default_http_exception_handler(Request(scope=scope), AttributeError("oops"))
    assert response.status_code == HTTP_500_INTERNAL_SERVER_ERROR
    assert response.content == {
        "detail": "Internal Server Error",
        "status_code": HTTP_500_INTERNAL_SERVER_ERROR,
    }


def test_exception_handler_middleware_exception_handlers_mapping() -> None:
    @get("/")
    def handler() -> None: ...

    def exception_handler(request: Request, exc: Exception) -> Response:
        return Response(content={"an": "error"}, status_code=HTTP_500_INTERNAL_SERVER_ERROR)

    app = Litestar(route_handlers=[handler], exception_handlers={Exception: exception_handler}, openapi_config=None)
    assert app.asgi_router.root_route_map_node.children["/"].asgi_handlers["GET"][0].exception_handlers == {  # type: ignore[attr-defined]
        Exception: exception_handler,
        StarletteHTTPException: _starlette_exception_handler,
    }


def test_exception_handler_middleware_calls_app_level_after_exception_hook() -> None:
    @get("/test")
    def handler() -> None:
        raise RuntimeError()

    async def after_exception_hook_handler(exc: Exception, scope: "Scope") -> None:
        app = scope.get("app")
        assert isinstance(exc, RuntimeError)
        assert app
        assert not app.state.called
        app.state.called = True

    with create_test_client(handler, after_exception=[after_exception_hook_handler]) as client:
        setattr(client.app.state, "called", False)
        assert not client.app.state.called
        response = client.get("/test")
        assert response.status_code == HTTP_500_INTERNAL_SERVER_ERROR
        assert client.app.state.called


@pytest.mark.parametrize(
    "is_debug, logging_config, should_log",
    [
        (True, LoggingConfig(log_exceptions="debug"), True),
        (False, LoggingConfig(log_exceptions="debug"), False),
        (True, LoggingConfig(log_exceptions="always"), True),
        (False, LoggingConfig(log_exceptions="always"), True),
        (True, LoggingConfig(log_exceptions="never"), False),
        (False, LoggingConfig(log_exceptions="never"), False),
        (True, None, False),
        (False, None, False),
    ],
)
def test_exception_handler_default_logging(
    get_logger: "GetLogger",
    caplog: "LogCaptureFixture",
    is_debug: bool,
    logging_config: Optional[LoggingConfig],
    should_log: bool,
) -> None:
    @get("/test")
    def handler() -> None:
        raise ValueError("Test debug exception")

    app = Litestar([handler], logging_config=logging_config, debug=is_debug)

    with caplog.at_level("ERROR", "litestar"), TestClient(app=app) as client:
        client.app.logger = get_logger("litestar")
        response = client.get("/test")
        assert response.status_code == HTTP_500_INTERNAL_SERVER_ERROR
        if is_debug:
            assert "Test debug exception" in response.text
        else:
            assert "Internal Server Error" in response.text

        if should_log:
            assert len(caplog.records) == 1
            assert caplog.records[0].levelname == "ERROR"
            assert caplog.records[0].message.startswith(
                "exception raised on http connection to route /test\n\nTraceback (most recent call last):\n"
            )
        else:
            assert not caplog.records
            assert "exception raised on http connection request to route /test" not in response.text


@pytest.mark.parametrize(
    "is_debug, logging_config, should_log",
    [
        (True, StructLoggingConfig(log_exceptions="debug"), True),
        (False, StructLoggingConfig(log_exceptions="debug"), False),
        (True, StructLoggingConfig(log_exceptions="always"), True),
        (False, StructLoggingConfig(log_exceptions="always"), True),
        (True, StructLoggingConfig(log_exceptions="never"), False),
        (False, StructLoggingConfig(log_exceptions="never"), False),
        (True, None, False),
        (False, None, False),
    ],
)
def test_exception_handler_struct_logging(
    get_logger: "GetLogger",
    capsys: CaptureFixture,
    is_debug: bool,
    logging_config: Optional[LoggingConfig],
    should_log: bool,
) -> None:
    @get("/test")
    def handler() -> None:
        raise ValueError("Test debug exception")

    app = Litestar([handler], logging_config=logging_config, debug=is_debug)

    with TestClient(app=app) as client, capture_logs() as cap_logs:
        response = client.get("/test")
        assert response.status_code == HTTP_500_INTERNAL_SERVER_ERROR
        if is_debug:
            assert "Test debug exception" in response.text
        else:
            assert "Internal Server Error" in response.text

        if should_log:
            assert len(cap_logs) == 1
            assert cap_logs[0].get("connection_type") == "http"
            assert cap_logs[0].get("path") == "/test"
            assert cap_logs[0].get("traceback")
            assert cap_logs[0].get("event") == "Uncaught Exception"
            assert cap_logs[0].get("log_level") == "error"
        else:
            assert not cap_logs


def test_traceback_truncate_default_logging(
    get_logger: "GetLogger",
    caplog: "LogCaptureFixture",
) -> None:
    @get("/test")
    def handler() -> None:
        raise ValueError("Test debug exception")

    app = Litestar([handler], logging_config=LoggingConfig(log_exceptions="always", traceback_line_limit=1))

    with caplog.at_level("ERROR", "litestar"), TestClient(app=app) as client:
        client.app.logger = get_logger("litestar")
        response = client.get("/test")
        assert response.status_code == HTTP_500_INTERNAL_SERVER_ERROR
        assert "Internal Server Error" in response.text

        assert len(caplog.records) == 1
        assert caplog.records[0].levelname == "ERROR"
        assert caplog.records[0].message == (
            "exception raised on http connection to route /test\n\nTraceback (most recent call last):\nValueError: Test debug exception\n"
        )


def test_traceback_truncate_struct_logging() -> None:
    @get("/test")
    def handler() -> None:
        raise ValueError("Test debug exception")

    app = Litestar([handler], logging_config=StructLoggingConfig(log_exceptions="always", traceback_line_limit=1))

    with TestClient(app=app) as client, capture_logs() as cap_logs:
        response = client.get("/test")
        assert response.status_code == HTTP_500_INTERNAL_SERVER_ERROR
        assert len(cap_logs) == 1
        assert cap_logs[0].get("traceback") == "ValueError: Test debug exception\n"


def handler(_: Any, __: Any) -> Any:
    return None


def handler_2(_: Any, __: Any) -> Any:
    return None


@pytest.mark.parametrize(
    ["mapping", "exc", "expected"],
    [
        ({}, Exception, None),
        ({HTTP_400_BAD_REQUEST: handler}, ValidationException(), handler),
        ({InternalServerException: handler}, InternalServerException(), handler),
        ({HTTP_500_INTERNAL_SERVER_ERROR: handler}, Exception(), handler),
        ({TypeError: handler}, TypeError(), handler),
        ({Exception: handler}, ValidationException(), handler),
        ({ValueError: handler}, ValidationException(), handler),
        ({ValidationException: handler}, Exception(), None),
        ({HTTP_500_INTERNAL_SERVER_ERROR: handler}, ValidationException(), None),
        ({HTTP_500_INTERNAL_SERVER_ERROR: handler, HTTPException: handler_2}, ValidationException(), handler_2),
        ({HTTPException: handler, ValidationException: handler_2}, ValidationException(), handler_2),
        ({HTTPException: handler, ValidationException: handler_2}, InternalServerException(), handler),
        ({HTTP_500_INTERNAL_SERVER_ERROR: handler, HTTPException: handler_2}, InternalServerException(), handler),
    ],
)
def test_get_exception_handler(mapping: ExceptionHandlersMap, exc: Exception, expected: Any) -> None:
    assert get_exception_handler(mapping, exc) == expected


@pytest.mark.filterwarnings("ignore::litestar.utils.warnings.LitestarWarning:")
def test_pdb_on_exception(mocker: MockerFixture) -> None:
    @get("/test")
    def handler() -> None:
        raise ValueError("Test debug exception")

    mock_post_mortem = mocker.patch("litestar.middleware.exceptions.middleware.pdb.post_mortem")

    app = Litestar([handler], pdb_on_exception=True)

    with TestClient(app=app) as client:
        response = client.get("/test")

    assert response.status_code == HTTP_500_INTERNAL_SERVER_ERROR
    mock_post_mortem.assert_called_once()


def test_get_debug_from_scope(get_logger: "GetLogger", caplog: "LogCaptureFixture") -> None:
    @get("/test")
    def handler() -> None:
        raise ValueError("Test debug exception")

    app = Litestar([handler], debug=False)
    app.debug = True

    with caplog.at_level("ERROR", "litestar"), TestClient(app=app) as client:
        client.app.logger = get_logger("litestar")
        response = client.get("/test")

        assert response.status_code == HTTP_500_INTERNAL_SERVER_ERROR
        assert "Test debug exception" in response.text
        assert len(caplog.records) == 1
        assert caplog.records[0].levelname == "ERROR"
        assert caplog.records[0].message.startswith(
            "exception raised on http connection to route /test\n\nTraceback (most recent call last):\n"
        )


def test_get_symbol_name_where_type_doesnt_support_bool() -> None:
    class Test:
        def __bool__(self) -> bool:
            raise TypeError("This type doesn't support bool")

        def method(self) -> None:
            raise RuntimeError("Oh no!")

    exc = None

    try:
        Test().method()
    except Exception as e:
        exc = e

    if exc is not None and exc.__traceback__ is not None:
        frame = getinnerframes(exc.__traceback__, 2)[-1]
        assert get_symbol_name(frame) == "Test.method"


@pytest.mark.parametrize("media_type", list(MediaType))
def test_serialize_custom_types(media_type: MediaType) -> None:
    # ensure type encoders are passed down to the created response so custom types that
    # might end up as part of a ValidationException are handled properly
    # https://github.com/litestar-org/litestar/issues/2867
    # https://github.com/litestar-org/litestar/issues/3192
    class Foo:
        def __init__(self, value: str) -> None:
            self.value = value

    @get(media_type=media_type)
    def handler() -> None:
        raise ValidationException(extra={"foo": Foo("bar")})

    with create_test_client([handler], type_encoders={Foo: lambda f: f.value}) as client:
        res = client.get("/")
        assert res.json()["extra"] == {"foo": "bar"}



from typing import Dict, List, Optional, Union

from pydantic import BaseModel
from pydantic.dataclasses import dataclass as pydantic_dataclass
from pydantic.v1 import BaseModel as BaseModelV1
from pydantic.v1.dataclasses import dataclass as dataclass_v1

from tests.models import DataclassPet


@pydantic_dataclass
class PydanticDataclassPerson:
    first_name: str
    last_name: str
    id: str
    optional: Optional[str]
    complex: Dict[str, List[Dict[str, str]]]
    union: Union[int, List[str]]
    pets: Optional[List[DataclassPet]] = None


class PydanticPerson(BaseModel):
    first_name: str
    last_name: str
    id: str
    optional: Optional[str]
    complex: Dict[str, List[Dict[str, str]]]
    union: Union[int, List[str]]
    pets: Optional[List[DataclassPet]] = None


class PydanticV1Person(BaseModelV1):
    first_name: str
    last_name: str
    id: str
    optional: Optional[str]
    complex: Dict[str, List[Dict[str, str]]]
    union: Union[int, List[str]]
    pets: Optional[List[DataclassPet]] = None


@dataclass_v1
class PydanticV1DataclassPerson:
    first_name: str
    last_name: str
    id: str
    optional: Optional[str]
    complex: Dict[str, List[Dict[str, str]]]
    union: Union[int, List[str]]
    pets: Optional[List[DataclassPet]] = None



from __future__ import annotations

import re
from abc import ABC, abstractmethod
from datetime import date, datetime, time, timedelta
from decimal import Decimal
from pathlib import Path
from typing import TYPE_CHECKING, Any, Callable
from uuid import UUID

import msgspec

from litestar.exceptions import ImproperlyConfiguredException
from litestar.types.internal_types import PathParameterDefinition
from litestar.utils import join_paths, normalize_path

if TYPE_CHECKING:
    from litestar.enums import ScopeType
    from litestar.types import Method, Receive, Scope, Send


def _parse_datetime(value: str) -> datetime:
    return msgspec.convert(value, datetime)


def _parse_date(value: str) -> date:
    return msgspec.convert(value, date)


def _parse_time(value: str) -> time:
    return msgspec.convert(value, time)


def _parse_timedelta(value: str) -> timedelta:
    try:
        return msgspec.convert(value, timedelta)
    except msgspec.ValidationError:
        return timedelta(seconds=int(float(value)))


param_match_regex = re.compile(r"{(.*?)}")

param_type_map = {
    "str": str,
    "int": int,
    "float": float,
    "uuid": UUID,
    "decimal": Decimal,
    "date": date,
    "datetime": datetime,
    "time": time,
    "timedelta": timedelta,
    "path": Path,
}


parsers_map: dict[Any, Callable[[Any], Any]] = {
    float: float,
    int: int,
    Decimal: Decimal,
    UUID: UUID,
    date: _parse_date,
    datetime: _parse_datetime,
    time: _parse_time,
    timedelta: _parse_timedelta,
}


class BaseRoute(ABC):
    """Base Route class used by Litestar.

    It's an abstract class meant to be extended.
    """

    __slots__ = (
        "app",
        "handler_names",
        "methods",
        "path",
        "path_format",
        "path_parameters",
        "path_components",
        "scope_type",
    )

    def __init__(
        self,
        *,
        handler_names: list[str],
        path: str,
        scope_type: ScopeType,
        methods: list[Method] | None = None,
    ) -> None:
        """Initialize the route.

        Args:
            handler_names: Names of the associated handler functions
            path: Base path of the route
            scope_type: Type of the ASGI scope
            methods: Supported methods
        """
        self.path, self.path_format, self.path_components, self.path_parameters = self._parse_path(path)
        self.handler_names = handler_names
        self.scope_type = scope_type
        self.methods = set(methods or [])

    @abstractmethod
    async def handle(self, scope: Scope, receive: Receive, send: Send) -> None:
        """ASGI App of the route.

        Args:
            scope: The ASGI connection scope.
            receive: The ASGI receive function.
            send: The ASGI send function.

        Returns:
            None
        """
        raise NotImplementedError("Route subclasses must implement handle which serves as the ASGI app entry point")

    @staticmethod
    def _validate_path_parameter(param: str, path: str) -> None:
        """Validate that a path parameter adheres to the required format and datatypes.

        Raises:
            ImproperlyConfiguredException: If the parameter has an invalid format.
        """
        if len(param.split(":")) != 2:
            raise ImproperlyConfiguredException(
                f"Path parameters should be declared with a type using the following pattern: '{{parameter_name:type}}', e.g. '/my-path/{{my_param:int}}' in path: '{path}'"
            )
        param_name, param_type = (p.strip() for p in param.split(":"))
        if not param_name:
            raise ImproperlyConfiguredException("Path parameter names should be of length greater than zero")
        if param_type not in param_type_map:
            raise ImproperlyConfiguredException(
                f"Path parameters should be declared with an allowed type, i.e. one of {', '.join(param_type_map.keys())} in path: '{path}'"
            )

    @classmethod
    def _parse_path(
        cls, path: str
    ) -> tuple[str, str, list[str | PathParameterDefinition], dict[str, PathParameterDefinition]]:
        """Normalize and parse a path.

        Splits the path into a list of components, parsing any that are path parameters. Also builds the OpenAPI
        compatible path, which does not include the type of the path parameters.

        Returns:
            A 3-tuple of the normalized path, the OpenAPI formatted path, and the list of parsed components.
        """
        path = normalize_path(path)

        parsed_components: list[str | PathParameterDefinition] = []
        path_format_components = []
        path_parameters: dict[str, PathParameterDefinition] = {}

        components = [component for component in path.split("/") if component]
        for component in components:
            if param_match := param_match_regex.fullmatch(component):
                param = param_match.group(1)
                cls._validate_path_parameter(param, path)
                param_name, param_type = (p.strip() for p in param.split(":"))
                type_class = param_type_map[param_type]
                parser = parsers_map[type_class] if type_class not in {str, Path} else None
                if param_name in path_parameters:
                    raise ImproperlyConfiguredException(f"Duplicate parameter '{param_name}' detected in '{path}'.")
                param_definition = PathParameterDefinition(name=param_name, type=type_class, full=param, parser=parser)
                parsed_components.append(param_definition)
                path_parameters[param_name] = param_definition
                path_format_components.append("{" + param_name + "}")
            else:
                parsed_components.append(component)
                path_format_components.append(component)

        path_format = join_paths(path_format_components)

        return path, path_format, parsed_components, path_parameters



from __future__ import annotations

from typing import TYPE_CHECKING

import pytest
from click import Group

from litestar import Litestar, MediaType, get
from litestar.constants import UNDEFINED_SENTINELS
from litestar.contrib.attrs import AttrsSchemaPlugin
from litestar.contrib.pydantic import PydanticDIPlugin, PydanticInitPlugin, PydanticPlugin, PydanticSchemaPlugin
from litestar.contrib.sqlalchemy.plugins import SQLAlchemySerializationPlugin
from litestar.plugins import CLIPluginProtocol, InitPluginProtocol, OpenAPISchemaPlugin, PluginRegistry
from litestar.plugins.core import MsgspecDIPlugin
from litestar.testing import create_test_client
from litestar.typing import FieldDefinition

if TYPE_CHECKING:
    from litestar.config.app import AppConfig


def test_plugin_on_app_init() -> None:
    @get("/", media_type=MediaType.TEXT)
    def greet() -> str:
        return "hello world"

    tag = "on_app_init_called"

    def on_startup(app: Litestar) -> None:
        app.state.called = True

    class PluginWithInitOnly(InitPluginProtocol):
        def on_app_init(self, app_config: AppConfig) -> AppConfig:
            app_config.tags.append(tag)
            app_config.on_startup.append(on_startup)
            app_config.route_handlers.append(greet)
            return app_config

    with create_test_client(plugins=[PluginWithInitOnly()]) as client:
        response = client.get("/")
        assert response.text == "hello world"

        assert tag in client.app.tags
        assert client.app.state.called


def test_plugin_registry() -> None:
    class CLIPlugin(CLIPluginProtocol):
        def on_cli_init(self, cli: Group) -> None:
            pass

    cli_plugin = CLIPlugin()
    serialization_plugin = SQLAlchemySerializationPlugin()
    openapi_plugin = PydanticSchemaPlugin()
    init_plugin = PydanticInitPlugin()

    registry = PluginRegistry([cli_plugin, serialization_plugin, openapi_plugin, init_plugin])

    assert registry.openapi == (openapi_plugin,)
    assert registry.cli == (cli_plugin,)
    assert registry.serialization == (serialization_plugin,)
    assert registry.init == (init_plugin,)

    assert openapi_plugin in registry
    assert serialization_plugin in registry
    assert init_plugin in registry
    assert cli_plugin in registry

    assert set(registry) == {openapi_plugin, cli_plugin, init_plugin, serialization_plugin}


def test_plugin_registry_get() -> None:
    class CLIPlugin(CLIPluginProtocol):
        def on_cli_init(self, cli: Group) -> None:
            pass

    cli_plugin = CLIPlugin()

    with pytest.raises(KeyError, match="No plugin of type 'CLIPlugin' registered"):
        PluginRegistry([]).get(CLIPlugin)

    assert PluginRegistry([cli_plugin]).get(CLIPlugin) is cli_plugin


def test_plugin_registry_stringified_get() -> None:
    class CLIPlugin(CLIPluginProtocol):
        def on_cli_init(self, cli: Group) -> None:
            pass

    cli_plugin = CLIPlugin()
    pydantic_plugin = PydanticPlugin()
    with pytest.raises(KeyError):
        PluginRegistry([CLIPlugin()]).get(
            "litestar2.contrib.pydantic.PydanticPlugin"
        )  # not a fqdn.  should fail # type: ignore[list-item]
        PluginRegistry([]).get("CLIPlugin")  # not a fqdn.  should fail # type: ignore[list-item]

    assert PluginRegistry([cli_plugin, pydantic_plugin]).get(CLIPlugin) is cli_plugin
    assert PluginRegistry([cli_plugin, pydantic_plugin]).get(PydanticPlugin) is pydantic_plugin
    assert PluginRegistry([cli_plugin, pydantic_plugin]).get("PydanticPlugin") is pydantic_plugin
    assert (
        PluginRegistry([cli_plugin, pydantic_plugin]).get("litestar.contrib.pydantic.PydanticPlugin") is pydantic_plugin
    )


def test_openapi_schema_plugin_is_constrained_field() -> None:
    assert OpenAPISchemaPlugin.is_constrained_field(FieldDefinition.from_annotation(str)) is False


def test_openapi_schema_plugin_is_undefined_sentinel() -> None:
    for value in UNDEFINED_SENTINELS:
        assert OpenAPISchemaPlugin.is_undefined_sentinel(value) is False


@pytest.mark.parametrize(("init_plugin",), [(PydanticInitPlugin(),), (None,)])
@pytest.mark.parametrize(("schema_plugin",), [(PydanticSchemaPlugin(),), (None,)])
@pytest.mark.parametrize(("attrs_plugin",), [(AttrsSchemaPlugin(),), (None,)])
def test_app_get_default_plugins(
    init_plugin: PydanticInitPlugin, schema_plugin: PydanticSchemaPlugin, attrs_plugin: AttrsSchemaPlugin
) -> None:
    plugins = [p for p in (init_plugin, schema_plugin, attrs_plugin) if p is not None]
    any_pydantic = bool(init_plugin) or bool(schema_plugin)
    default_plugins = Litestar._get_default_plugins(plugins)  # type: ignore[arg-type]
    if not any_pydantic:
        assert {type(p) for p in default_plugins} == {
            PydanticPlugin,
            AttrsSchemaPlugin,
            PydanticDIPlugin,
            MsgspecDIPlugin,
        }
    else:
        assert {type(p) for p in default_plugins} == {
            PydanticInitPlugin,
            PydanticSchemaPlugin,
            AttrsSchemaPlugin,
            PydanticDIPlugin,
            MsgspecDIPlugin,
        }



import pytest
from docs.examples.templating.returning_templates_jinja import app as jinja_app
from docs.examples.templating.returning_templates_mako import app as mako_app
from docs.examples.templating.returning_templates_minijinja import app as minijinja_app

from litestar.testing import TestClient

apps_with_expected_responses = [
    (jinja_app, "Jinja", "Hello <strong>Jinja</strong>", "Hello <strong>Jinja</strong> using strings"),
    (mako_app, "Mako", "Hello <strong>Mako</strong>", "Hello <strong>Mako</strong> using strings"),
    (minijinja_app, "Minijinja", "Hello <strong>Minijinja</strong>", "Hello <strong>Minijinja</strong> using strings"),
]


@pytest.mark.parametrize("app, app_name, file_response, string_response", apps_with_expected_responses)
@pytest.mark.parametrize("template_type", ["file", "string"])
def test_returning_templates(app, app_name, file_response, string_response, template_type):
    with TestClient(app) as client:
        response = client.get(f"/{template_type}", params={"name": app_name})
        if template_type == "file":
            assert response.text.strip() == file_response
        elif template_type == "string":
            assert response.text.strip() == string_response



from litestar.static_files.base import StaticFiles
from litestar.static_files.config import StaticFilesConfig, create_static_files_router

__all__ = ("StaticFiles", "StaticFilesConfig", "create_static_files_router")



from litestar import Controller, MediaType, asgi
from litestar.enums import ScopeType
from litestar.response.base import ASGIResponse
from litestar.status_codes import HTTP_200_OK
from litestar.testing import create_test_client
from litestar.types import Receive, Scope, Send


def test_handle_asgi() -> None:
    @asgi(path="/")
    async def root_asgi_handler(scope: Scope, receive: Receive, send: Send) -> None:
        assert scope["type"] == ScopeType.HTTP
        assert scope["method"] == "GET"
        response = ASGIResponse(body=b"Hello World", media_type=MediaType.TEXT)
        await response(scope, receive, send)

    class MyController(Controller):
        path = "/asgi"

        @asgi()
        async def root_asgi_handler(self, scope: Scope, receive: Receive, send: Send) -> None:
            assert scope["type"] == ScopeType.HTTP
            assert scope["method"] == "GET"
            response = ASGIResponse(body=b"Hello World", media_type=MediaType.TEXT)
            await response(scope, receive, send)

    with create_test_client([root_asgi_handler, MyController]) as client:
        response = client.get("/")
        assert response.status_code == HTTP_200_OK
        assert response.text == "Hello World"
        response = client.get("/asgi")
        assert response.status_code == HTTP_200_OK
        assert response.text == "Hello World"


def test_asgi_signature_namespace() -> None:
    class MyController(Controller):
        path = "/asgi"
        signature_namespace = {"b": Receive}

        @asgi(signature_namespace={"c": Send})
        async def root_asgi_handler(
            self,
            scope: "a",  # type:ignore[name-defined]  # noqa: F821
            receive: "b",  # type:ignore[name-defined]  # noqa: F821
            send: "c",  # type:ignore[name-defined]  # noqa: F821
        ) -> None:
            await ASGIResponse(body=scope["path"].encode(), media_type=MediaType.TEXT)(scope, receive, send)

    with create_test_client([MyController], signature_namespace={"a": Scope}) as client:
        response = client.get("/asgi")
        assert response.status_code == HTTP_200_OK
        assert response.text == "/asgi"



from __future__ import annotations

import asyncio

from litestar.channels import Subscriber


async def get_from_stream(subscriber: Subscriber, count: int) -> list[bytes]:
    async def getter() -> list[bytes]:
        items = []
        async for item in subscriber.iter_events():
            items.append(item)
            if len(items) == count:
                break
        return items

    return await asyncio.wait_for(getter(), timeout=1)



from __future__ import annotations

from dataclasses import dataclass
from typing import TYPE_CHECKING

from litestar.openapi.spec.base import BaseSchemaObject

if TYPE_CHECKING:
    from litestar.openapi.spec.media_type import OpenAPIMediaType

__all__ = ("RequestBody",)


@dataclass
class RequestBody(BaseSchemaObject):
    """Describes a single request body."""

    content: dict[str, OpenAPIMediaType]
    """
    **REQUIRED**. The content of the request body.
    The key is a media type or `media type range <https://tools.ietf.org/html/rfc7231#appendix-D>`_ and the value
    describes it.

    For requests that match multiple keys, only the most specific key is applicable. e.g. ``text/plain`` overrides
    ``text/*``
    """

    description: str | None = None
    """A brief description of the request body. This could contain examples of use.

    `CommonMark syntax <https://spec.commonmark.org/>`_ MAY be used for rich text representation.
    """

    required: bool = False
    """Determines if the request body is required in the request.

    Defaults to ``False``.
    """



from dataclasses import dataclass
from typing import Any, Callable, Dict, List, Type
from unittest.mock import ANY, MagicMock

import pytest
from typing_extensions import Annotated

from litestar import Controller, Litestar, get, post
from litestar._openapi.datastructures import OpenAPIContext
from litestar._openapi.request_body import create_request_body
from litestar.datastructures.upload_file import UploadFile
from litestar.dto import AbstractDTO
from litestar.enums import RequestEncodingType
from litestar.handlers import BaseRouteHandler
from litestar.openapi.config import OpenAPIConfig
from litestar.openapi.spec import RequestBody
from litestar.params import Body
from litestar.typing import FieldDefinition


@dataclass
class FormData:
    cv: UploadFile
    image: UploadFile


RequestBodyFactory = Callable[[BaseRouteHandler, FieldDefinition], RequestBody]


@pytest.fixture()
def openapi_context() -> OpenAPIContext:
    return OpenAPIContext(
        openapi_config=OpenAPIConfig(title="test", version="1.0.0", create_examples=True),
        plugins=[],
    )


@pytest.fixture()
def create_request(openapi_context: OpenAPIContext) -> RequestBodyFactory:
    def _factory(route_handler: BaseRouteHandler, data_field: FieldDefinition) -> RequestBody:
        return create_request_body(
            context=openapi_context,
            handler_id=route_handler.handler_id,
            resolved_data_dto=route_handler.resolve_data_dto(),
            data_field=data_field,
        )

    return _factory


def test_create_request_body(person_controller: Type[Controller], create_request: RequestBodyFactory) -> None:
    for route in Litestar(route_handlers=[person_controller]).routes:
        for route_handler, _ in route.route_handler_map.values():  # type: ignore[union-attr]
            handler_fields = route_handler.parsed_fn_signature.parameters
            if "data" in handler_fields:
                request_body = create_request(route_handler, handler_fields["data"])
                assert request_body


def test_request_body_schema_extra() -> None:
    @dataclass
    class RequestBody:
        foo: str

    @get()
    async def handler(
        body1: Annotated[
            RequestBody,
            Body(
                title="Default title",
                schema_extra={
                    "title": "Overridden title",
                },
            ),
        ],
    ) -> Any:
        return body1

    app = Litestar([handler])
    schema = app.openapi_schema.to_schema()
    resp = next(iter(schema["components"]["schemas"].values()))
    assert resp["title"] == "Overridden title"


def test_upload_single_file_schema_generation() -> None:
    @post(path="/file-upload")
    async def handle_file_upload(
        data: UploadFile = Body(media_type=RequestEncodingType.MULTI_PART),
    ) -> None:
        return None

    app = Litestar([handle_file_upload])
    schema = app.openapi_schema.to_schema()

    assert schema["paths"]["/file-upload"]["post"]["requestBody"]["content"]["multipart/form-data"]["schema"] == {
        "properties": {"file": {"type": "string", "format": "binary", "contentMediaType": "application/octet-stream"}},
        "type": "object",
    }


def test_upload_list_of_files_schema_generation() -> None:
    @post(path="/file-list-upload")
    async def handle_file_list_upload(
        data: List[UploadFile] = Body(media_type=RequestEncodingType.MULTI_PART),
    ) -> None:
        return None

    app = Litestar([handle_file_list_upload])
    schema = app.openapi_schema.to_schema()

    assert schema["paths"]["/file-list-upload"]["post"]["requestBody"]["content"]["multipart/form-data"]["schema"] == {
        "type": "object",
        "properties": {
            "files": {
                "items": {"type": "string", "contentMediaType": "application/octet-stream", "format": "binary"},
                "type": "array",
            }
        },
    }


def test_upload_file_dict_schema_generation() -> None:
    @post(path="/file-dict-upload")
    async def handle_file_list_upload(
        data: Dict[str, UploadFile] = Body(media_type=RequestEncodingType.MULTI_PART),
    ) -> None:
        return None

    app = Litestar([handle_file_list_upload])
    schema = app.openapi_schema.to_schema()

    assert schema["paths"]["/file-dict-upload"]["post"]["requestBody"]["content"]["multipart/form-data"]["schema"] == {
        "type": "object",
        "properties": {
            "files": {
                "items": {"type": "string", "contentMediaType": "application/octet-stream", "format": "binary"},
                "type": "array",
            }
        },
    }


def test_upload_file_model_schema_generation() -> None:
    @post(path="/form-upload")
    async def handle_form_upload(
        data: FormData = Body(media_type=RequestEncodingType.MULTI_PART),
    ) -> None:
        return None

    app = Litestar([handle_form_upload])
    schema = app.openapi_schema.to_schema()

    assert schema["paths"]["/form-upload"]["post"]["requestBody"]["content"]["multipart/form-data"] == {
        "schema": {"$ref": "#/components/schemas/FormData"}
    }
    assert schema["components"] == {
        "schemas": {
            "FormData": {
                "properties": {
                    "cv": {"type": "string", "contentMediaType": "application/octet-stream", "format": "binary"},
                    "image": {"type": "string", "contentMediaType": "application/octet-stream", "format": "binary"},
                },
                "type": "object",
                "required": ["cv", "image"],
                "title": "FormData",
            }
        }
    }


def test_request_body_generation_with_dto(create_request: RequestBodyFactory) -> None:
    mock_dto = MagicMock(spec=AbstractDTO)

    @post(path="/form-upload", dto=mock_dto)  # pyright: ignore
    async def handler(data: Dict[str, Any]) -> None:
        return None

    Litestar(route_handlers=[handler])
    field_definition = FieldDefinition.from_annotation(Dict[str, Any])
    create_request(handler, field_definition)
    mock_dto.create_openapi_schema.assert_called_once_with(
        field_definition=field_definition, handler_id=handler.handler_id, schema_creator=ANY
    )



from __future__ import annotations

from collections import defaultdict
from typing import TYPE_CHECKING, Iterator, Sequence

from litestar.exceptions import ImproperlyConfiguredException
from litestar.openapi.spec import Reference, Schema

if TYPE_CHECKING:
    from litestar.openapi import OpenAPIConfig
    from litestar.plugins import OpenAPISchemaPluginProtocol


class RegisteredSchema:
    """Object to store a schema and any references to it."""

    def __init__(self, key: tuple[str, ...], schema: Schema, references: list[Reference]) -> None:
        """Create a new RegisteredSchema object.

        Args:
            key: The key used to register the schema.
            schema: The schema object.
            references: A list of references to the schema.
        """
        self.key = key
        self.schema = schema
        self.references = references


class SchemaRegistry:
    """A registry for object schemas.

    This class is used to store schemas that we reference from other parts of the spec.

    Its main purpose is to allow us to generate the components/schemas section of the spec once we have
    collected all the schemas that should be included.

    This allows us to determine a path to the schema in the components/schemas section of the spec that
    is unique and as short as possible.
    """

    def __init__(self) -> None:
        self._schema_key_map: dict[tuple[str, ...], RegisteredSchema] = {}
        self._schema_reference_map: dict[int, RegisteredSchema] = {}
        self._model_name_groups: defaultdict[str, list[RegisteredSchema]] = defaultdict(list)

    def get_schema_for_key(self, key: tuple[str, ...]) -> Schema:
        """Get a registered schema by its key.

        Args:
            key: The key to the schema to get.

        Returns:
            A RegisteredSchema object.
        """
        if key not in self._schema_key_map:
            self._schema_key_map[key] = registered_schema = RegisteredSchema(key, Schema(), [])
            self._model_name_groups[key[-1]].append(registered_schema)
        return self._schema_key_map[key].schema

    def get_reference_for_key(self, key: tuple[str, ...]) -> Reference | None:
        """Get a reference to a registered schema by its key.

        Args:
            key: The key to the schema to get.

        Returns:
            A Reference object.
        """
        if key not in self._schema_key_map:
            return None
        registered_schema = self._schema_key_map[key]
        reference = Reference(f"#/components/schemas/{'_'.join(key)}")
        registered_schema.references.append(reference)
        self._schema_reference_map[id(reference)] = registered_schema
        return reference

    def from_reference(self, reference: Reference) -> RegisteredSchema:
        """Get a registered schema by its reference.

        Args:
            reference: The reference to the schema to get.

        Returns:
            A RegisteredSchema object.
        """
        return self._schema_reference_map[id(reference)]

    def __iter__(self) -> Iterator[RegisteredSchema]:
        """Iterate over the registered schemas."""
        return iter(self._schema_key_map.values())

    @staticmethod
    def set_reference_paths(name: str, registered_schema: RegisteredSchema) -> None:
        """Set the reference paths for a registered schema."""
        for reference in registered_schema.references:
            reference.ref = f"#/components/schemas/{name}"

    @staticmethod
    def remove_common_prefix(tuples: list[tuple[str, ...]]) -> list[tuple[str, ...]]:
        """Remove the common prefix from a list of tuples.

        Args:
            tuples: A list of tuples to remove the common prefix from.

        Returns:
            A list of tuples with the common prefix removed.
        """

        def longest_common_prefix(tuples_: list[tuple[str, ...]]) -> tuple[str, ...]:
            """Find the longest common prefix of a list of tuples.

            Args:
                tuples_: A list of tuples to find the longest common prefix of.

            Returns:
                The longest common prefix of the tuples.
            """
            prefix_ = tuples_[0]
            for t in tuples_:
                # Compare the current prefix with each tuple and shorten it
                prefix_ = prefix_[: min(len(prefix_), len(t))]
                for i in range(len(prefix_)):
                    if prefix_[i] != t[i]:
                        prefix_ = prefix_[:i]
                        break
            return prefix_

        prefix = longest_common_prefix(tuples)
        prefix_length = len(prefix)
        return [t[prefix_length:] for t in tuples]

    def generate_components_schemas(self) -> dict[str, Schema]:
        """Generate the components/schemas section of the spec.

        Returns:
            A dictionary of schemas.
        """
        components_schemas: dict[str, Schema] = {}

        for name, name_group in self._model_name_groups.items():
            if len(name_group) == 1:
                self.set_reference_paths(name, name_group[0])
                components_schemas[name] = name_group[0].schema
                continue

            full_keys = [registered_schema.key for registered_schema in name_group]
            names = ["_".join(k) for k in self.remove_common_prefix(full_keys)]
            for name_, registered_schema in zip(names, name_group):
                self.set_reference_paths(name_, registered_schema)
                components_schemas[name_] = registered_schema.schema

        # Sort them by name to ensure they're always generated in the same order.
        return {name: components_schemas[name] for name in sorted(components_schemas.keys())}


class OpenAPIContext:
    def __init__(
        self,
        openapi_config: OpenAPIConfig,
        plugins: Sequence[OpenAPISchemaPluginProtocol],
    ) -> None:
        self.openapi_config = openapi_config
        self.plugins = plugins
        self.operation_ids: set[str] = set()
        self.schema_registry = SchemaRegistry()

    def add_operation_id(self, operation_id: str) -> None:
        """Add an operation ID to the context.

        Args:
            operation_id: Operation ID to add.
        """
        if operation_id in self.operation_ids:
            raise ImproperlyConfiguredException(
                "operation_ids must be unique, "
                f"please ensure the value of 'operation_id' is either not set or unique for {operation_id}"
            )
        self.operation_ids.add(operation_id)



from typing import Any, Optional

from litestar import MediaType, get
from litestar.contrib.htmx._utils import HTMXHeaders
from litestar.contrib.htmx.request import HTMXRequest
from litestar.status_codes import HTTP_200_OK
from litestar.testing import create_test_client


def test_health_check() -> None:
    @get("/health-check", media_type=MediaType.TEXT)
    def health_check() -> str:
        return "healthy"

    with create_test_client(route_handlers=health_check) as client:
        response = client.get("/health-check")
        assert response.status_code == HTTP_200_OK
        assert response.text == "healthy"


async def test_bool_default() -> None:
    @get("/")
    def handler(request: HTMXRequest) -> bool:
        return bool(request.htmx)

    with create_test_client(route_handlers=[handler], request_class=HTMXRequest) as client:
        response = client.get("/")
        assert response.text == "false"


async def test_bool_false() -> None:
    @get("/")
    def handler(request: HTMXRequest) -> bool:
        return bool(request.htmx)

    with create_test_client(route_handlers=[handler], request_class=HTMXRequest) as client:
        response = client.get("/", headers={HTMXHeaders.REQUEST.value: "false"})
        assert response.text == "false"


async def test_bool_true() -> None:
    @get("/")
    def handler(request: HTMXRequest) -> bool:
        return bool(request.htmx)

    with create_test_client(route_handlers=[handler], request_class=HTMXRequest) as client:
        response = client.get("/", headers={HTMXHeaders.REQUEST.value: "true"})
        assert response.text == "true"


async def test_boosted_default() -> None:
    @get("/")
    def handler(request: HTMXRequest) -> bool:
        return request.htmx.boosted

    with create_test_client(route_handlers=[handler], request_class=HTMXRequest) as client:
        response = client.get("/")
        assert response.text == "false"


async def test_boosted_set() -> None:
    @get("/")
    def handler(request: HTMXRequest) -> bool:
        return request.htmx.boosted

    with create_test_client(route_handlers=[handler], request_class=HTMXRequest) as client:
        response = client.get("/", headers={HTMXHeaders.BOOSTED.value: "true"})
        assert response.text == "true"


def test_current_url_default() -> None:
    @get("/")
    def handler(request: HTMXRequest) -> Optional[str]:
        assert request.htmx.current_url is None
        return request.htmx.current_url

    with create_test_client(route_handlers=[handler], request_class=HTMXRequest) as client:
        response = client.get("/")
        assert response.text == "null"


def test_current_url_set() -> None:
    @get("/")
    def handler(request: HTMXRequest) -> Optional[str]:
        assert request.htmx.current_url == "https://example.com"
        return request.htmx.current_url

    with create_test_client(route_handlers=[handler], request_class=HTMXRequest) as client:
        response = client.get("/", headers={HTMXHeaders.CURRENT_URL.value: "https://example.com"})
        assert response.text == "https://example.com"


def test_current_url_set_url_encoded() -> None:
    @get("/")
    def handler(request: HTMXRequest) -> Optional[str]:
        return request.htmx.current_url

    with create_test_client(route_handlers=[handler], request_class=HTMXRequest) as client:
        response = client.get(
            "/",
            headers={
                HTMXHeaders.CURRENT_URL.value: "https%3A%2F%2Fexample.com%2F%3F",
                f"{HTMXHeaders.CURRENT_URL.value}-URI-AutoEncoded": "true",
            },
        )
        assert response.text == "https://example.com/?"


def test_current_url_abs_path_default() -> None:
    @get("/")
    def handler(request: HTMXRequest) -> Optional[str]:
        assert request.htmx.current_url_abs_path is None
        return request.htmx.current_url_abs_path

    with create_test_client(route_handlers=[handler], request_class=HTMXRequest) as client:
        response = client.get("/")
        assert response.text == "null"


def test_current_url_abs_path_set() -> None:
    @get("/")
    def handler(request: HTMXRequest) -> Optional[str]:
        assert request.htmx.current_url_abs_path == "/duck/?quack=true#h2"
        return request.htmx.current_url_abs_path

    with create_test_client(route_handlers=[handler], request_class=HTMXRequest) as client:
        response = client.get(
            "/", headers={HTMXHeaders.CURRENT_URL.value: "http://testserver.local/duck/?quack=true#h2"}
        )
        assert response.text == "/duck/?quack=true#h2"


def test_current_url_abs_path_set_other_domain() -> None:
    @get("/")
    def handler(request: HTMXRequest) -> Optional[str]:
        assert request.htmx.current_url_abs_path is None
        return request.htmx.current_url_abs_path

    with create_test_client(route_handlers=[handler], request_class=HTMXRequest) as client:
        response = client.get("/", headers={HTMXHeaders.CURRENT_URL.value: "http://example.com/duck/?quack=true#h2"})
        assert response.text == "null"


def test_history_restore_request_false() -> None:
    @get("/")
    def handler(request: HTMXRequest) -> bool:
        assert request.htmx.history_restore_request is False
        return request.htmx.history_restore_request

    with create_test_client(route_handlers=[handler], request_class=HTMXRequest) as client:
        response = client.get("/", headers={HTMXHeaders.HISTORY_RESTORE_REQUEST.value: "false"})
        assert response.text == "false"


def test_history_restore_request_true() -> None:
    @get("/")
    def handler(request: HTMXRequest) -> bool:
        assert request.htmx.history_restore_request is True
        return request.htmx.history_restore_request

    with create_test_client(route_handlers=[handler], request_class=HTMXRequest) as client:
        response = client.get("/", headers={HTMXHeaders.HISTORY_RESTORE_REQUEST.value: "true"})
        assert response.text == "true"


def test_prompt_default() -> None:
    @get("/")
    def handler(request: HTMXRequest) -> Optional[str]:
        assert request.htmx.prompt is None
        return request.htmx.prompt

    with create_test_client(route_handlers=[handler], request_class=HTMXRequest) as client:
        response = client.get("/")
        assert response.text == "null"


def test_prompt_set() -> None:
    @get("/")
    def handler(request: HTMXRequest) -> Optional[str]:
        assert request.htmx.prompt == "Yes"
        return request.htmx.prompt

    with create_test_client(route_handlers=[handler], request_class=HTMXRequest) as client:
        response = client.get("/", headers={HTMXHeaders.PROMPT.value: "Yes"})
        assert response.text == "Yes"


def test_target_default() -> None:
    @get("/")
    def handler(request: HTMXRequest) -> Optional[str]:
        assert request.htmx.target is None
        return request.htmx.target

    with create_test_client(route_handlers=[handler], request_class=HTMXRequest) as client:
        response = client.get("/")
        assert response.text == "null"


def test_target_set() -> None:
    @get("/")
    def handler(request: HTMXRequest) -> Optional[str]:
        assert request.htmx.target == "#element"
        return request.htmx.target

    with create_test_client(route_handlers=[handler], request_class=HTMXRequest) as client:
        response = client.get("/", headers={HTMXHeaders.TARGET.value: "#element"})
        assert response.text == "#element"


def test_trigger_default() -> None:
    @get("/")
    def handler(request: HTMXRequest) -> Optional[str]:
        assert request.htmx.trigger is None
        return request.htmx.trigger

    with create_test_client(route_handlers=[handler], request_class=HTMXRequest) as client:
        response = client.get("/")
        assert response.text == "null"


def test_trigger_set() -> None:
    @get("/")
    def handler(request: HTMXRequest) -> Optional[str]:
        assert request.htmx.trigger == "#element"
        return request.htmx.trigger

    with create_test_client(route_handlers=[handler], request_class=HTMXRequest) as client:
        response = client.get("/", headers={HTMXHeaders.TRIGGER_ID.value: "#element"})
        assert response.text == "#element"


def test_trigger_name_default() -> None:
    @get("/")
    def handler(request: HTMXRequest) -> Optional[str]:
        assert request.htmx.trigger_name is None
        return request.htmx.trigger_name

    with create_test_client(route_handlers=[handler], request_class=HTMXRequest) as client:
        response = client.get("/")
        assert response.text == "null"


def test_trigger_name_set() -> None:
    @get("/")
    def handler(request: HTMXRequest) -> Optional[str]:
        assert request.htmx.trigger_name == "name_of_element"
        return request.htmx.trigger_name

    with create_test_client(route_handlers=[handler], request_class=HTMXRequest) as client:
        response = client.get("/", headers={HTMXHeaders.TRIGGER_NAME.value: "name_of_element"})
        assert response.text == "name_of_element"


def test_triggering_event_none() -> None:
    @get("/")
    def handler(request: HTMXRequest) -> None:
        assert request.htmx.triggering_event is None
        return request.htmx.triggering_event

    with create_test_client(route_handlers=[handler], request_class=HTMXRequest) as client:
        response = client.get("/")
        assert response.text == "null"


def test_triggering_event_bad_json() -> None:
    @get("/")
    def handler(request: HTMXRequest) -> None:
        assert request.htmx.triggering_event is None
        return request.htmx.triggering_event

    with create_test_client(route_handlers=[handler], request_class=HTMXRequest) as client:
        response = client.get("/", headers={HTMXHeaders.TRIGGERING_EVENT.value: "{"})
        assert response.text == "null"


def test_triggering_event_good_json() -> None:
    @get("/")
    def handler(request: HTMXRequest) -> Any:
        return request.htmx.triggering_event

    with create_test_client(route_handlers=[handler], request_class=HTMXRequest) as client:
        response = client.get(
            "/",
            headers={
                HTMXHeaders.TRIGGERING_EVENT.value: "%7B%22target%22%3A%20null%7D",
                f"{HTMXHeaders.TRIGGERING_EVENT.value}-uri-autoencoded": "true",
            },
        )
        assert response.text == '{"target":null}'



import html
from os import urandom
from pathlib import Path
from typing import Any, Optional

import pytest
from bs4 import BeautifulSoup

from litestar import MediaType, WebSocket, delete, get, patch, post, put, websocket
from litestar.config.csrf import CSRFConfig
from litestar.contrib.jinja import JinjaTemplateEngine
from litestar.contrib.mako import MakoTemplateEngine
from litestar.enums import RequestEncodingType
from litestar.handlers import HTTPRouteHandler
from litestar.params import Body
from litestar.response.template import Template
from litestar.status_codes import HTTP_200_OK, HTTP_201_CREATED, HTTP_403_FORBIDDEN
from litestar.template.config import TemplateConfig
from litestar.testing import create_test_client


def handler_fn() -> None:
    pass


@pytest.fixture
def get_handler() -> HTTPRouteHandler:
    return get()(handler_fn)


@pytest.fixture
def post_handler() -> HTTPRouteHandler:
    return post()(handler_fn)


@pytest.fixture
def put_handler() -> HTTPRouteHandler:
    return put()(handler_fn)


@pytest.fixture
def delete_handler() -> HTTPRouteHandler:
    return delete()(handler_fn)


@pytest.fixture
def patch_handler() -> HTTPRouteHandler:
    return patch()(handler_fn)


def test_csrf_successful_flow(get_handler: HTTPRouteHandler, post_handler: HTTPRouteHandler) -> None:
    with create_test_client(
        route_handlers=[get_handler, post_handler], csrf_config=CSRFConfig(secret="secret")
    ) as client:
        response = client.get("/")
        assert response.status_code == HTTP_200_OK

        csrf_token: Optional[str] = response.cookies.get("csrftoken")
        assert csrf_token is not None

        set_cookie_header = response.headers.get("set-cookie")
        assert set_cookie_header is not None
        assert set_cookie_header.split("; ") == [
            f"csrftoken={csrf_token}",
            "Path=/",
            "SameSite=lax",
        ]

        response = client.post("/", headers={"x-csrftoken": csrf_token})
        assert response.status_code == HTTP_201_CREATED


@pytest.mark.parametrize(
    "method",
    ["POST", "PUT", "DELETE", "PATCH"],
)
def test_unsafe_method_fails_without_csrf_header(
    method: str,
    get_handler: HTTPRouteHandler,
    post_handler: HTTPRouteHandler,
    put_handler: HTTPRouteHandler,
    delete_handler: HTTPRouteHandler,
    patch_handler: HTTPRouteHandler,
) -> None:
    with create_test_client(
        route_handlers=[get_handler, post_handler, put_handler, delete_handler, patch_handler],
        csrf_config=CSRFConfig(secret="secret"),
    ) as client:
        response = client.get("/")
        assert response.status_code == HTTP_200_OK

        csrf_token: Optional[str] = response.cookies.get("csrftoken")
        assert csrf_token is not None

        response = client.request(method, "/")
        assert response.status_code == HTTP_403_FORBIDDEN
        assert response.json() == {"detail": "CSRF token verification failed", "status_code": 403}


def test_invalid_csrf_token(get_handler: HTTPRouteHandler, post_handler: HTTPRouteHandler) -> None:
    with create_test_client(
        route_handlers=[get_handler, post_handler], csrf_config=CSRFConfig(secret="secret")
    ) as client:
        response = client.get("/")
        assert response.status_code == HTTP_200_OK

        csrf_token: Optional[str] = response.cookies.get("csrftoken")
        assert csrf_token is not None

        response = client.post("/", headers={"x-csrftoken": f"{csrf_token}invalid"})
        assert response.status_code == HTTP_403_FORBIDDEN
        assert response.json() == {"detail": "CSRF token verification failed", "status_code": 403}


def test_csrf_token_too_short(get_handler: HTTPRouteHandler, post_handler: HTTPRouteHandler) -> None:
    with create_test_client(
        route_handlers=[get_handler, post_handler], csrf_config=CSRFConfig(secret="secret")
    ) as client:
        response = client.get("/")
        assert response.status_code == HTTP_200_OK

        assert "csrftoken" in response.cookies

        response = client.post("/", headers={"x-csrftoken": "too-short"})
        assert response.status_code == HTTP_403_FORBIDDEN
        assert response.json() == {"detail": "CSRF token verification failed", "status_code": 403}


def test_websocket_ignored() -> None:
    @websocket(path="/")
    async def websocket_handler(socket: WebSocket) -> None:
        await socket.accept()
        await socket.send_json({"data": "123"})
        await socket.close()

    with create_test_client(
        route_handlers=[websocket_handler], csrf_config=CSRFConfig(secret="secret")
    ) as client, client.websocket_connect("/") as ws:
        response = ws.receive_json()
        assert response is not None


def test_custom_csrf_config(get_handler: HTTPRouteHandler, post_handler: HTTPRouteHandler) -> None:
    with create_test_client(
        base_url="http://test.com",
        route_handlers=[get_handler, post_handler],
        csrf_config=CSRFConfig(
            secret="secret",
            cookie_name="custom-csrftoken",
            header_name="x-custom-csrftoken",
        ),
    ) as client:
        response = client.get("/")
        assert response.status_code == HTTP_200_OK

        csrf_token: Optional[str] = response.cookies.get("custom-csrftoken")
        assert csrf_token is not None

        set_cookie_header = response.headers.get("set-cookie")
        assert set_cookie_header is not None
        assert set_cookie_header.split("; ") == [
            f"custom-csrftoken={csrf_token}",
            "Path=/",
            "SameSite=lax",
        ]

        response = client.post("/", headers={"x-custom-csrftoken": csrf_token})
        assert response.status_code == HTTP_201_CREATED


@pytest.mark.parametrize(
    "engine, template",
    (
        (JinjaTemplateEngine, "{{csrf_input}}"),
        (MakoTemplateEngine, "${csrf_input}"),
    ),
)
def test_csrf_form_parsing(engine: Any, template: str, tmp_path: Path) -> None:
    @get(path="/", media_type=MediaType.HTML)
    def handler() -> Template:
        return Template(template_name="abc.html")

    @post("/")
    def form_handler(data: dict = Body(media_type=RequestEncodingType.URL_ENCODED)) -> dict:
        return data

    with create_test_client(
        route_handlers=[handler, form_handler],
        template_config=TemplateConfig(
            directory=tmp_path,
            engine=engine,
        ),
        csrf_config=CSRFConfig(secret=str(urandom(10))),
    ) as client:
        url = f"{client.base_url!s}/"
        Path(tmp_path / "abc.html").write_text(
            f'<html><body><div><form action="{url}" method="post">{template}</form></div></body></html>'
        )
        _ = client.get("/")
        response = client.get("/")
        html_soup = BeautifulSoup(html.unescape(response.text), features="html.parser")
        data = {"_csrf_token": html_soup.body.div.form.input.attrs.get("value")}  # type: ignore[union-attr]
        response = client.post("/", data=data)
        assert response.status_code == HTTP_201_CREATED
        assert response.json() == data


def test_csrf_middleware_exclude_from_check_via_opts() -> None:
    @post("/", exclude_from_csrf=True)
    def post_handler(data: dict = Body(media_type=RequestEncodingType.URL_ENCODED)) -> dict:
        return data

    with create_test_client(
        route_handlers=[post_handler],
        csrf_config=CSRFConfig(secret=str(urandom(10))),
    ) as client:
        data = {"field": "value"}
        response = client.post("/", data=data)
        assert response.status_code == HTTP_201_CREATED
        assert response.json() == data


def test_csrf_middleware_exclude_from_check() -> None:
    @post("/protected-handler")
    def post_handler(data: dict = Body(media_type=RequestEncodingType.URL_ENCODED)) -> dict:
        return data

    @post("/unprotected-handler")
    def post_handler2(data: dict = Body(media_type=RequestEncodingType.URL_ENCODED)) -> dict:
        return data

    with create_test_client(
        route_handlers=[post_handler, post_handler2],
        csrf_config=CSRFConfig(secret=str(urandom(10)), exclude=["unprotected-handler"]),
    ) as client:
        data = {"field": "value"}
        response = client.post("/protected-handler", data=data)
        assert response.status_code == HTTP_403_FORBIDDEN

        response = client.post("/unprotected-handler", data=data)
        assert response.status_code == HTTP_201_CREATED
        assert response.json() == data


def test_csrf_middleware_configure_name_for_exclude_from_check_via_opts() -> None:
    @post("/handler", exclude_from_csrf=True)
    def post_handler(data: dict = Body(media_type=RequestEncodingType.URL_ENCODED)) -> dict:
        return data

    @post("/handler2", custom_exclude_from_csrf=True)
    def post_handler2(data: dict = Body(media_type=RequestEncodingType.URL_ENCODED)) -> dict:
        return data

    with create_test_client(
        route_handlers=[post_handler, post_handler2],
        csrf_config=CSRFConfig(secret=str(urandom(10)), exclude_from_csrf_key="custom_exclude_from_csrf"),
    ) as client:
        data = {"field": "value"}
        response = client.post("/handler", data=data)
        assert response.status_code == HTTP_403_FORBIDDEN

        data = {"field": "value"}
        response = client.post("/handler2", data=data)
        assert response.status_code == HTTP_201_CREATED
        assert response.json() == data



from typing import Generic, Optional, TypeVar

from attrs import define
from typing_extensions import Annotated

from litestar.contrib.attrs.attrs_schema_plugin import AttrsSchemaPlugin
from litestar.openapi.spec import OpenAPIType
from litestar.openapi.spec.schema import Schema
from litestar.typing import FieldDefinition
from litestar.utils.helpers import get_name
from tests.helpers import get_schema_for_field_definition

T = TypeVar("T")


@define
class AttrsGeneric(Generic[T]):
    foo: T
    optional_foo: Optional[T]
    annotated_foo: Annotated[T, object()]


def test_schema_generation_with_generic_classes() -> None:
    cls = AttrsGeneric[int]
    expected_foo_schema = Schema(type=OpenAPIType.INTEGER)
    expected_optional_foo_schema = Schema(one_of=[Schema(type=OpenAPIType.NULL), Schema(type=OpenAPIType.INTEGER)])

    field_definition = FieldDefinition.from_kwarg(name=get_name(cls), annotation=cls)
    properties = get_schema_for_field_definition(field_definition, plugins=[AttrsSchemaPlugin()]).properties

    assert properties
    assert properties["foo"] == expected_foo_schema
    assert properties["annotated_foo"] == expected_foo_schema
    assert properties["optional_foo"] == expected_optional_foo_schema



from __future__ import annotations

import functools
from pathlib import Path
from typing import TYPE_CHECKING, Any, Mapping, Protocol, TypeVar, cast

from typing_extensions import ParamSpec

from litestar.exceptions import ImproperlyConfiguredException, MissingDependencyException, TemplateNotFoundException
from litestar.template.base import (
    TemplateCallableType,
    TemplateEngineProtocol,
    TemplateProtocol,
    csrf_token,
    url_for,
    url_for_static_asset,
)
from litestar.utils.deprecation import warn_deprecation

try:
    from minijinja import Environment  # type:ignore[import-untyped]
    from minijinja import TemplateError as MiniJinjaTemplateNotFound
except ImportError as e:
    raise MissingDependencyException("minijinja") from e

if TYPE_CHECKING:
    from typing import Callable

    C = TypeVar("C", bound="Callable")

    def pass_state(func: C) -> C: ...

else:
    from minijinja import pass_state

__all__ = (
    "MiniJinjaTemplateEngine",
    "StateProtocol",
)

P = ParamSpec("P")
T = TypeVar("T")


class StateProtocol(Protocol):
    auto_escape: str | None
    current_block: str | None
    env: Environment
    name: str

    def lookup(self, key: str) -> Any | None: ...


def _transform_state(func: TemplateCallableType[Mapping[str, Any], P, T]) -> TemplateCallableType[StateProtocol, P, T]:
    """Transform a template callable to receive a ``StateProtocol`` instance as first argument.

    This is for wrapping callables like ``url_for()`` that receive a mapping as first argument so they can be used
    with minijinja which passes a ``StateProtocol`` instance as first argument.
    """

    @functools.wraps(func)
    @pass_state
    def wrapped(state: StateProtocol, /, *args: P.args, **kwargs: P.kwargs) -> T:
        template_context = {"request": state.lookup("request"), "csrf_input": state.lookup("csrf_input")}
        return func(template_context, *args, **kwargs)

    return wrapped


class MiniJinjaTemplate(TemplateProtocol):
    """Initialize a template.

    Args:
        template: Base ``MiniJinjaTemplate`` used by the underlying minijinja engine
    """

    def __init__(self, engine: Environment, template_name: str) -> None:
        super().__init__()
        self.engine = engine
        self.template_name = template_name

    def render(self, *args: Any, **kwargs: Any) -> str:
        """Render a template.

        Args:
            args: Positional arguments passed to the engines ``render`` function
            kwargs: Keyword arguments passed to the engines ``render`` function

        Returns:
            Rendered template as a string
        """
        try:
            return str(self.engine.render_template(self.template_name, *args, **kwargs))
        except MiniJinjaTemplateNotFound as err:
            raise TemplateNotFoundException(template_name=self.template_name) from err


class MiniJinjaTemplateEngine(TemplateEngineProtocol["MiniJinjaTemplate", StateProtocol]):
    """The engine instance."""

    def __init__(self, directory: Path | list[Path] | None = None, engine_instance: Environment | None = None) -> None:
        """Minijinja based TemplateEngine.

        Args:
            directory: Direct path or list of directory paths from which to serve templates.
            engine_instance: A Minijinja Environment instance.
        """
        super().__init__(directory, engine_instance)
        if directory and engine_instance:
            raise ImproperlyConfiguredException(
                "You must provide either a directory or a minijinja Environment instance."
            )
        if directory:

            def _loader(name: str) -> str:
                """Load a template from a directory.

                Args:
                    name: The name of the template

                Returns:
                    The template as a string

                Raises:
                    TemplateNotFoundException: if no template is found.
                """
                directories = directory if isinstance(directory, list) else [directory]

                for d in directories:
                    template_path = Path(d) / name  # pyright: ignore[reportGeneralTypeIssues]
                    if template_path.exists():
                        return template_path.read_text()
                raise TemplateNotFoundException(template_name=name)

            self.engine = Environment(loader=_loader)
        elif engine_instance:
            self.engine = engine_instance
        else:
            raise ImproperlyConfiguredException(
                "You must provide either a directory or a minijinja Environment instance."
            )

        self.register_template_callable("url_for", _transform_state(url_for))
        self.register_template_callable("csrf_token", _transform_state(csrf_token))
        self.register_template_callable("url_for_static_asset", _transform_state(url_for_static_asset))

    def get_template(self, template_name: str) -> MiniJinjaTemplate:
        """Retrieve a template by matching its name (dotted path) with files in the directory or directories provided.

        Args:
            template_name: A dotted path

        Returns:
            MiniJinjaTemplate instance

        Raises:
            TemplateNotFoundException: if no template is found.
        """
        return MiniJinjaTemplate(self.engine, template_name)

    def register_template_callable(
        self,
        key: str,
        template_callable: TemplateCallableType[StateProtocol, P, T],
    ) -> None:
        """Register a callable on the template engine.

        Args:
            key: The callable key, i.e. the value to use inside the template to call the callable.
            template_callable: A callable to register.

        Returns:
            None
        """

        def is_decorated(func: Callable) -> bool:
            return hasattr(func, "__wrapped__") or func.__name__ not in globals()

        if not is_decorated(template_callable):
            template_callable = _transform_state(template_callable)  # type: ignore[arg-type] # pragma: no cover
        self.engine.add_global(key, pass_state(template_callable))

    def render_string(self, template_string: str, context: Mapping[str, Any]) -> str:
        """Render a template from a string with the given context.

        Args:
            template_string: The template string to render.
            context: A dictionary of variables to pass to the template.

        Returns:
            The rendered template as a string.
        """
        return self.engine.render_str(template_string, **context)  # type: ignore[no-any-return]

    @classmethod
    def from_environment(cls, minijinja_environment: Environment) -> MiniJinjaTemplateEngine:
        """Create a MiniJinjaTemplateEngine from an existing minijinja Environment instance.

        Args:
            minijinja_environment (Environment): A minijinja Environment instance.

        Returns:
            MiniJinjaTemplateEngine instance
        """
        return cls(directory=None, engine_instance=minijinja_environment)


@pass_state
def _minijinja_from_state(func: Callable, state: StateProtocol, *args: Any, **kwargs: Any) -> str:  # pragma: no cover
    template_context = {"request": state.lookup("request"), "csrf_input": state.lookup("csrf_input")}
    return cast(str, func(template_context, *args, **kwargs))


def __getattr__(name: str) -> Any:
    if name == "minijinja_from_state":
        warn_deprecation(
            "2.3.0",
            "minijinja_from_state",
            "import",
            removal_in="3.0.0",
            alternative="Use a callable that receives the minijinja State object as first argument.",
        )
        return _minijinja_from_state
    raise AttributeError(f"module {__name__!r} has no attribute {name!r}")



from typing import Type

from polyfactory import BaseFactory

from litestar import Controller, Litestar
from litestar._openapi.typescript_converter.converter import (
    convert_openapi_to_typescript,
)


def test_openapi_to_typescript_converter(person_controller: Type[Controller], pet_controller: Type[Controller]) -> None:
    BaseFactory.seed_random(1)
    app = Litestar(route_handlers=[person_controller, pet_controller])
    assert app.openapi_schema

    result = convert_openapi_to_typescript(openapi_schema=app.openapi_schema)
    assert (
        result.write()
        == """export namespace API {
	export namespace PetOwnerOrPetGetPetsOrOwners {
	export namespace Http200 {
	export type ResponseBody = ({
	age: number;
	name: string;
	species?: "Cat" | "Dog" | "Monkey" | "Pig";
} | {
	complex: {
	
};
	first_name: string;
	id: string;
	last_name: string;
	optional?: null | string;
	pets?: null | {
	age: number;
	name: string;
	species?: "Cat" | "Dog" | "Monkey" | "Pig";
}[];
})[];

	export interface ResponseHeaders {
	"x-my-tag"?: string;
};
};

	export namespace Http406 {
	export type ResponseBody = {
	detail: string;
	extra?: Record<string, unknown> | null | unknown[];
	status_code: number;
};
};
};

	export namespace PetPets {
	export namespace Http200 {
	export type ResponseBody = {
	age: number;
	name: string;
	species?: "Cat" | "Dog" | "Monkey" | "Pig";
}[];
};
};

	export namespace ServiceIdPersonBulkBulkCreatePerson {
	export interface HeaderParameters {
	secret: string;
};

	export namespace Http201 {
	export type ResponseBody = {
	complex: {
	
};
	first_name: string;
	id: string;
	last_name: string;
	optional: null | string;
	pets: null | {
	age: number;
	name: string;
	species: "Cat" | "Dog" | "Monkey" | "Pig";
}[];
}[];
};

	export namespace Http400 {
	export type ResponseBody = {
	detail: string;
	extra?: Record<string, unknown> | null | unknown[];
	status_code: number;
};
};

	export interface PathParameters {
	service_id: number;
};

	export type RequestBody = {
	complex: {
	
};
	first_name: string;
	id: string;
	last_name: string;
	optional: null | string;
	pets: null | {
	age: number;
	name: string;
	species: "Cat" | "Dog" | "Monkey" | "Pig";
}[];
}[];
};

	export namespace ServiceIdPersonBulkBulkPartialUpdatePerson {
	export interface HeaderParameters {
	secret: string;
};

	export namespace Http200 {
	export type ResponseBody = {
	complex: {
	
};
	first_name: string;
	id: string;
	last_name: string;
	optional: null | string;
	pets: null | {
	age: number;
	name: string;
	species: "Cat" | "Dog" | "Monkey" | "Pig";
}[];
}[];
};

	export namespace Http400 {
	export type ResponseBody = {
	detail: string;
	extra?: Record<string, unknown> | null | unknown[];
	status_code: number;
};
};

	export interface PathParameters {
	service_id: number;
};

	export type RequestBody = {
	complex: {
	
};
	first_name: string;
	id: string;
	last_name: string;
	optional: null | string;
	pets: null | {
	age: number;
	name: string;
	species: "Cat" | "Dog" | "Monkey" | "Pig";
}[];
}[];
};

	export namespace ServiceIdPersonBulkBulkUpdatePerson {
	export interface HeaderParameters {
	secret: string;
};

	export namespace Http200 {
	export type ResponseBody = {
	complex: {
	
};
	first_name: string;
	id: string;
	last_name: string;
	optional?: null | string;
	pets?: null | {
	age: number;
	name: string;
	species?: "Cat" | "Dog" | "Monkey" | "Pig";
}[];
}[];
};

	export namespace Http400 {
	export type ResponseBody = {
	detail: string;
	extra?: Record<string, unknown> | null | unknown[];
	status_code: number;
};
};

	export interface PathParameters {
	service_id: number;
};

	export type RequestBody = {
	complex: {
	
};
	first_name: string;
	id: string;
	last_name: string;
	optional?: null | string;
	pets?: null | {
	age: number;
	name: string;
	species?: "Cat" | "Dog" | "Monkey" | "Pig";
}[];
}[];
};

	export namespace ServiceIdPersonCreatePerson {
	export interface HeaderParameters {
	secret: string;
};

	export namespace Http201 {
	export type ResponseBody = {
	complex: {
	
};
	first_name: string;
	id: string;
	last_name: string;
	optional?: null | string;
	pets?: null | {
	age: number;
	name: string;
	species?: "Cat" | "Dog" | "Monkey" | "Pig";
}[];
};
};

	export namespace Http400 {
	export type ResponseBody = {
	detail: string;
	extra?: Record<string, unknown> | null | unknown[];
	status_code: number;
};
};

	export interface PathParameters {
	service_id: number;
};

	export type RequestBody = {
	complex: {
	
};
	first_name: string;
	id: string;
	last_name: string;
	optional?: null | string;
	pets?: null | {
	age: number;
	name: string;
	species?: "Cat" | "Dog" | "Monkey" | "Pig";
}[];
};
};

	export namespace ServiceIdPersonDataclassGetPersonDataclass {
	export namespace Http200 {
	export type ResponseBody = {
	complex: {
	
};
	first_name: string;
	id: string;
	last_name: string;
	optional?: null | string;
	pets?: null | {
	age: number;
	name: string;
	species?: "Cat" | "Dog" | "Monkey" | "Pig";
}[];
};
};

	export namespace Http400 {
	export type ResponseBody = {
	detail: string;
	extra?: Record<string, unknown> | null | unknown[];
	status_code: number;
};
};

	export interface PathParameters {
	service_id: number;
};
};

	export namespace ServiceIdPersonGetPersons {
	export interface CookieParameters {
	value: number;
};

	export interface HeaderParameters {
	secret: string;
};

	export namespace Http200 {
	export type ResponseBody = {
	complex: {
	
};
	first_name: string;
	id: string;
	last_name: string;
	optional?: null | string;
	pets?: null | {
	age: number;
	name: string;
	species?: "Cat" | "Dog" | "Monkey" | "Pig";
}[];
}[];
};

	export namespace Http400 {
	export type ResponseBody = {
	detail: string;
	extra?: Record<string, unknown> | null | unknown[];
	status_code: number;
};
};

	export interface PathParameters {
	service_id: number;
};

	export interface QueryParameters {
	from_date?: null | number | string | string;
	gender?: "A" | "F" | "M" | "O" | ("A" | "F" | "M" | "O")[] | null;
	name?: null | string | string[];
	page: number;
	pageSize: number;
	to_date?: null | number | string | string;
};
};

	export namespace ServiceIdPersonPersonIdDeletePerson {
	export namespace Http204 {
	export type ResponseBody = undefined;
};

	export namespace Http400 {
	export type ResponseBody = {
	detail: string;
	extra?: Record<string, unknown> | null | unknown[];
	status_code: number;
};
};

	export interface PathParameters {
	person_id: string;
	service_id: number;
};
};

	export namespace ServiceIdPersonPersonIdGetPersonById {
	export namespace Http200 {
	export type ResponseBody = {
	complex: {
	
};
	first_name: string;
	id: string;
	last_name: string;
	optional?: null | string;
	pets?: null | {
	age: number;
	name: string;
	species?: "Cat" | "Dog" | "Monkey" | "Pig";
}[];
};
};

	export namespace Http400 {
	export type ResponseBody = {
	detail: string;
	extra?: Record<string, unknown> | null | unknown[];
	status_code: number;
};
};

	export interface PathParameters {
	person_id: string;
	service_id: number;
};
};

	export namespace ServiceIdPersonPersonIdPartialUpdatePerson {
	export namespace Http200 {
	export type ResponseBody = {
	complex: {
	
};
	first_name: string;
	id: string;
	last_name: string;
	optional: null | string;
	pets: null | {
	age: number;
	name: string;
	species: "Cat" | "Dog" | "Monkey" | "Pig";
}[];
};
};

	export namespace Http400 {
	export type ResponseBody = {
	detail: string;
	extra?: Record<string, unknown> | null | unknown[];
	status_code: number;
};
};

	export interface PathParameters {
	person_id: string;
	service_id: number;
};

	export type RequestBody = {
	complex: {
	
};
	first_name: string;
	id: string;
	last_name: string;
	optional: null | string;
	pets: null | {
	age: number;
	name: string;
	species: "Cat" | "Dog" | "Monkey" | "Pig";
}[];
};
};

	export namespace ServiceIdPersonPersonIdUpdatePerson {
	export namespace Http200 {
	export type ResponseBody = {
	complex: {
	
};
	first_name: string;
	id: string;
	last_name: string;
	optional?: null | string;
	pets?: null | {
	age: number;
	name: string;
	species?: "Cat" | "Dog" | "Monkey" | "Pig";
}[];
};
};

	export namespace Http400 {
	export type ResponseBody = {
	detail: string;
	extra?: Record<string, unknown> | null | unknown[];
	status_code: number;
};
};

	export interface PathParameters {
	person_id: string;
	service_id: number;
};

	export type RequestBody = {
	complex: {
	
};
	first_name: string;
	id: string;
	last_name: string;
	optional?: null | string;
	pets?: null | {
	age: number;
	name: string;
	species?: "Cat" | "Dog" | "Monkey" | "Pig";
}[];
};
};
};"""
    )



from __future__ import annotations

from pathlib import Path

from click import Context, group, option, pass_context
from click import Path as ClickPath

from ._utils import LitestarEnv, LitestarExtensionGroup
from .commands import core, schema, sessions

__all__ = ("litestar_group",)


@group(cls=LitestarExtensionGroup, context_settings={"help_option_names": ["-h", "--help"]})
@option("--app", "app_path", help="Module path to a Litestar application")
@option(
    "--app-dir",
    help="Look for APP in the specified directory, by adding this to the PYTHONPATH. Defaults to the current working directory.",
    default=None,
    type=ClickPath(dir_okay=True, file_okay=False, path_type=Path),
    show_default=False,
)
@pass_context
def litestar_group(ctx: Context, app_path: str | None, app_dir: Path | None = None) -> None:
    """Litestar CLI."""
    if ctx.obj is None:  # env has not been loaded yet, so we can lazy load it
        ctx.obj = lambda: LitestarEnv.from_env(app_path, app_dir=app_dir)


# add sub commands here

litestar_group.add_command(core.info_command)
litestar_group.add_command(core.run_command)
litestar_group.add_command(core.routes_command)
litestar_group.add_command(core.version_command)
litestar_group.add_command(sessions.sessions_group)
litestar_group.add_command(schema.schema_group)



from __future__ import annotations

import sys
from dataclasses import dataclass, field, replace
from typing import ClassVar, List
from unittest.mock import ANY

import pytest
from typing_extensions import Annotated

from litestar.dto import DataclassDTO, DTOField
from litestar.dto.data_structures import DTOFieldDefinition
from litestar.typing import FieldDefinition


@dataclass
class Model:
    a: int
    b: str = field(default="b")
    c: List[int] = field(default_factory=list)  # noqa: UP006
    d: ClassVar[float] = 1.0


@pytest.fixture(name="dto_type")
def fx_dto_type() -> type[DataclassDTO[Model]]:
    return DataclassDTO[Model]


@pytest.mark.skipif(sys.version_info >= (3, 9), reason="generic builtin collection")
def test_dataclass_field_definitions_38(dto_type: type[DataclassDTO[Model]]) -> None:
    expected = [
        replace(
            DTOFieldDefinition.from_field_definition(
                field_definition=FieldDefinition.from_kwarg(
                    name="a",
                    annotation=int,
                ),
                default_factory=None,
                model_name=Model.__name__,
                dto_field=DTOField(),
            ),
            metadata=ANY,
            type_wrappers=ANY,
            raw=ANY,
        ),
        replace(
            DTOFieldDefinition.from_field_definition(
                field_definition=FieldDefinition.from_kwarg(name="b", annotation=str, default="b"),
                default_factory=None,
                model_name=Model.__name__,
                dto_field=DTOField(),
            ),
            metadata=ANY,
            type_wrappers=ANY,
            raw=ANY,
        ),
        replace(
            DTOFieldDefinition.from_field_definition(
                field_definition=FieldDefinition.from_kwarg(
                    name="c",
                    annotation=List[int],
                ),
                default_factory=list,
                model_name=Model.__name__,
                dto_field=DTOField(),
            ),
            metadata=ANY,
            type_wrappers=ANY,
            raw=ANY,
        ),
    ]
    for field_def, exp in zip(dto_type.generate_field_definitions(Model), expected):
        assert field_def == exp


def test_dataclass_field_definitions(dto_type: type[DataclassDTO[Model]]) -> None:
    expected = [
        replace(
            DTOFieldDefinition.from_field_definition(
                field_definition=FieldDefinition.from_kwarg(
                    name="a",
                    annotation=int,
                ),
                default_factory=None,
                model_name=Model.__name__,
                dto_field=DTOField(),
            ),
            metadata=ANY,
            type_wrappers=ANY,
            raw=ANY,
        ),
        replace(
            DTOFieldDefinition.from_field_definition(
                field_definition=FieldDefinition.from_kwarg(name="b", annotation=str, default="b"),
                default_factory=None,
                model_name=Model.__name__,
                dto_field=DTOField(),
            ),
            metadata=ANY,
            type_wrappers=ANY,
            raw=ANY,
        ),
        replace(
            DTOFieldDefinition.from_field_definition(
                field_definition=FieldDefinition.from_kwarg(
                    name="c",
                    annotation=List[int],
                ),
                default_factory=list,
                model_name=Model.__name__,
                dto_field=DTOField(),
            ),
            metadata=ANY,
            type_wrappers=ANY,
            raw=ANY,
        ),
    ]
    for field_def, exp in zip(dto_type.generate_field_definitions(Model), expected):
        assert field_def == exp


def test_dataclass_detect_nested(dto_type: type[DataclassDTO[Model]]) -> None:
    assert dto_type.detect_nested_field(FieldDefinition.from_annotation(Model)) is True
    assert dto_type.detect_nested_field(FieldDefinition.from_annotation(int)) is False


ReadOnlyInt = Annotated[int, DTOField("read-only")]


def test_dataclass_dto_annotated_dto_field() -> None:
    Annotated[int, DTOField("read-only")]

    @dataclass
    class Model:
        a: Annotated[int, DTOField("read-only")]
        b: ReadOnlyInt

    dto_type = DataclassDTO[Model]
    fields = list(dto_type.generate_field_definitions(Model))
    assert fields[0].dto_field == DTOField("read-only")
    assert fields[1].dto_field == DTOField("read-only")



from typing import TYPE_CHECKING, Any, cast

import pytest

from litestar import get
from litestar.config.allowed_hosts import AllowedHostsConfig
from litestar.exceptions import ImproperlyConfiguredException
from litestar.middleware import MiddlewareProtocol
from litestar.middleware.allowed_hosts import AllowedHostsMiddleware
from litestar.status_codes import HTTP_200_OK, HTTP_400_BAD_REQUEST
from litestar.testing import create_test_client

if TYPE_CHECKING:
    from litestar.types import Receive, Scope, Send


class DummyApp(MiddlewareProtocol):  # pyright: ignore
    async def __call__(self, scope: "Scope", receive: "Receive", send: "Send") -> None:
        return


def test_allowed_hosts_middleware() -> None:
    @get(path="/")
    def handler() -> None: ...

    client = create_test_client(route_handlers=[handler], allowed_hosts=["*.example.com", "moishe.zuchmir.com"])
    unpacked_middleware = []
    cur = client.app.asgi_router.root_route_map_node.children["/"].asgi_handlers["GET"][0]
    while hasattr(cur, "app"):
        unpacked_middleware.append(cur)
        cur = cast("Any", cur.app)
    unpacked_middleware.append(cur)

    assert len(unpacked_middleware) == 4
    allowed_hosts_middleware = cast("Any", unpacked_middleware[1])
    assert isinstance(allowed_hosts_middleware, AllowedHostsMiddleware)
    assert allowed_hosts_middleware.allowed_hosts_regex.pattern == ".*\\.example.com$|moishe.zuchmir.com"  # type: ignore[union-attr]


def test_allowed_hosts_middleware_hosts_regex() -> None:
    config = AllowedHostsConfig(allowed_hosts=["*.example.com", "moishe.zuchmir.com"])
    middleware = AllowedHostsMiddleware(app=DummyApp(), config=config)  # type: ignore[abstract]
    assert middleware.allowed_hosts_regex is not None
    assert middleware.allowed_hosts_regex.pattern == ".*\\.example.com$|moishe.zuchmir.com"

    assert middleware.allowed_hosts_regex.fullmatch("www.example.com")
    assert middleware.allowed_hosts_regex.fullmatch("other.example.com")
    assert middleware.allowed_hosts_regex.fullmatch("x.y.z.example.com")
    assert middleware.allowed_hosts_regex.fullmatch("moishe.zuchmir.com")

    assert not middleware.allowed_hosts_regex.fullmatch("www.example.x.com")
    assert not middleware.allowed_hosts_regex.fullmatch("josh.zuchmir.com")
    assert not middleware.allowed_hosts_regex.fullmatch("x.moishe.zuchmir.com")
    assert not middleware.allowed_hosts_regex.fullmatch("moishe.zuchmir.x.com")


def test_allowed_hosts_middleware_redirect_regex() -> None:
    config = AllowedHostsConfig(
        allowed_hosts=["*.example.com", "www.moishe.zuchmir.com", "www.yada.bada.bing.io", "example.com"]
    )
    middleware = AllowedHostsMiddleware(app=DummyApp(), config=config)  # type: ignore[abstract]
    assert middleware.redirect_domains is not None
    assert middleware.redirect_domains.pattern == "moishe.zuchmir.com|yada.bada.bing.io"

    assert middleware.redirect_domains.fullmatch("moishe.zuchmir.com")
    assert middleware.redirect_domains.fullmatch("yada.bada.bing.io")


def test_middleware_allowed_hosts() -> None:
    @get("/")
    def handler() -> dict:
        return {"hello": "world"}

    config = AllowedHostsConfig(allowed_hosts=["*.example.com", "moishe.zuchmir.com"])

    with create_test_client(handler, allowed_hosts=config) as client:
        client.base_url = "http://x.example.com"  # type: ignore[assignment]
        response = client.get("/")
        assert response.status_code == HTTP_200_OK

        client.base_url = "http://x.y.example.com"  # type: ignore[assignment]
        response = client.get("/")
        assert response.status_code == HTTP_200_OK

        client.base_url = "http://moishe.zuchmir.com"  # type: ignore[assignment]
        response = client.get("/")
        assert response.status_code == HTTP_200_OK

        client.base_url = "http://x.moishe.zuchmir.com"  # type: ignore[assignment]
        response = client.get("/")
        assert response.status_code == HTTP_400_BAD_REQUEST

        client.base_url = "http://x.example.x.com"  # type: ignore[assignment]
        response = client.get("/")
        assert response.status_code == HTTP_400_BAD_REQUEST


def test_middleware_allow_all() -> None:
    @get("/")
    def handler() -> dict:
        return {"hello": "world"}

    # contrived case - but if "*" is in hosts, we allow all.
    config = AllowedHostsConfig(allowed_hosts=["*", "*.example.com", "moishe.zuchmir.com"])

    with create_test_client(handler, allowed_hosts=config) as client:
        client.base_url = "http://any.domain.allowed.com"  # type: ignore[assignment]
        response = client.get("/")
        assert response.status_code == HTTP_200_OK


def test_middleware_redirect_on_www_by_default() -> None:
    @get("/")
    def handler() -> dict:
        return {"hello": "world"}

    config = AllowedHostsConfig(allowed_hosts=["www.moishe.zuchmir.com"])

    with create_test_client(handler, allowed_hosts=config) as client:
        client.base_url = "http://moishe.zuchmir.com"  # type: ignore[assignment]
        response = client.get("/")
        assert response.status_code == HTTP_200_OK
        assert str(response.url) == "http://www.moishe.zuchmir.com/"


def test_middleware_does_not_redirect_when_off() -> None:
    @get("/")
    def handler() -> dict:
        return {"hello": "world"}

    config = AllowedHostsConfig(allowed_hosts=["www.moishe.zuchmir.com"], www_redirect=False)

    with create_test_client(handler, allowed_hosts=config) as client:
        client.base_url = "http://moishe.zuchmir.com"  # type: ignore[assignment]
        response = client.get("/")
        assert response.status_code == HTTP_400_BAD_REQUEST


def test_validation_raises_for_wrong_wildcard_domain() -> None:
    with pytest.raises(ImproperlyConfiguredException):
        AllowedHostsConfig(allowed_hosts=["www.moishe.*.com"])



from pathlib import Path
from types import ModuleType
from typing import Callable, Dict

import pytest

from litestar import HttpMethod, Litestar, WebSocket, delete, get, route
from litestar.exceptions import ImproperlyConfiguredException, ValidationException
from litestar.handlers.http_handlers import HTTPRouteHandler
from litestar.response import File, Redirect
from litestar.status_codes import (
    HTTP_100_CONTINUE,
    HTTP_200_OK,
    HTTP_304_NOT_MODIFIED,
    HTTP_307_TEMPORARY_REDIRECT,
)
from tests.models import DataclassPerson


def test_route_handler_validation_http_method() -> None:
    # doesn't raise for http methods
    for value in (*list(HttpMethod), *[x.upper() for x in list(HttpMethod)]):
        assert route(http_method=value)  # type: ignore[arg-type, truthy-bool]

    # raises for invalid values
    with pytest.raises(ValidationException):
        HTTPRouteHandler(http_method="deleze")  # type: ignore[arg-type]

    # also when passing an empty list
    with pytest.raises(ImproperlyConfiguredException):
        route(http_method=[], status_code=HTTP_200_OK)

    # also when passing malformed tokens
    with pytest.raises(ValidationException):
        route(http_method=[HttpMethod.GET, "poft"], status_code=HTTP_200_OK)  # type: ignore[list-item]


async def test_function_validation() -> None:
    with pytest.raises(ImproperlyConfiguredException):

        @get(path="/")
        def method_with_no_annotation():  # type: ignore[no-untyped-def]
            pass

        Litestar(route_handlers=[method_with_no_annotation])

        method_with_no_annotation.on_registration(Litestar())

    with pytest.raises(ImproperlyConfiguredException):

        @delete(path="/")
        def method_with_no_content() -> Dict[str, str]:
            return {}

        Litestar(route_handlers=[method_with_no_content])

        method_with_no_content.on_registration(Litestar())

    with pytest.raises(ImproperlyConfiguredException):

        @get(path="/", status_code=HTTP_304_NOT_MODIFIED)
        def method_with_not_modified() -> Dict[str, str]:
            return {}

        Litestar(route_handlers=[method_with_not_modified])

        method_with_not_modified.on_registration(Litestar())

    with pytest.raises(ImproperlyConfiguredException):

        @get(path="/", status_code=HTTP_100_CONTINUE)
        def method_with_status_lower_than_200() -> Dict[str, str]:
            return {}

        Litestar(route_handlers=[method_with_status_lower_than_200])

        method_with_status_lower_than_200.on_registration(Litestar())

    @get(path="/", status_code=HTTP_307_TEMPORARY_REDIRECT)
    def redirect_method() -> Redirect:
        return Redirect("/test")

    Litestar(route_handlers=[redirect_method])

    redirect_method.on_registration(Litestar())

    @get(path="/")
    def file_method() -> File:
        return File(path=Path("."), filename="test_validations.py")

    Litestar(route_handlers=[file_method])

    file_method.on_registration(Litestar())

    assert not file_method.media_type

    with pytest.raises(ImproperlyConfiguredException):

        @get(path="/test")
        def test_function_1(socket: WebSocket) -> None:
            return None

        test_function_1.on_registration(Litestar())

    with pytest.raises(ImproperlyConfiguredException):

        @get("/person")
        def test_function_2(self, data: DataclassPerson) -> None:  # type: ignore[no-untyped-def]
            return None

        Litestar(route_handlers=[test_function_2])

        test_function_2.on_registration(Litestar())


@pytest.mark.parametrize(
    ("return_annotation", "should_raise"),
    [
        ("None", False),
        ("Response[None]", False),
        ("int", True),
        ("Response[int]", True),
        ("Response", True),
    ],
)
def test_204_response_annotations(
    return_annotation: str, should_raise: bool, create_module: Callable[[str], ModuleType]
) -> None:
    module = create_module(
        f"""
from litestar import get
from litestar.response import Response
from litestar.status_codes import HTTP_204_NO_CONTENT

@get(path="/", status_code=HTTP_204_NO_CONTENT)
def no_response_handler() -> {return_annotation}:
    pass
"""
    )

    if should_raise:
        with pytest.raises(ImproperlyConfiguredException):
            Litestar(route_handlers=[module.no_response_handler])
        return

    Litestar(route_handlers=[module.no_response_handler])



from __future__ import annotations

import secrets
from dataclasses import dataclass, field
from typing import TYPE_CHECKING, Any, Literal

from litestar.datastructures import Cookie, MutableScopeHeaders
from litestar.enums import ScopeType
from litestar.exceptions import ImproperlyConfiguredException
from litestar.middleware.session.base import ONE_DAY_IN_SECONDS, BaseBackendConfig, BaseSessionBackend
from litestar.types import Empty, Message, Scopes, ScopeSession
from litestar.utils.dataclass import extract_dataclass_items

__all__ = ("ServerSideSessionBackend", "ServerSideSessionConfig")


if TYPE_CHECKING:
    from litestar import Litestar
    from litestar.connection import ASGIConnection
    from litestar.stores.base import Store


class ServerSideSessionBackend(BaseSessionBackend["ServerSideSessionConfig"]):
    """Base class for server-side backends.

    Implements :class:`BaseSessionBackend` and defines and interface which subclasses can
    implement to facilitate the storage of session data.
    """

    def __init__(self, config: ServerSideSessionConfig) -> None:
        """Initialize ``ServerSideSessionBackend``

        Args:
            config: A subclass of ``ServerSideSessionConfig``
        """
        super().__init__(config=config)

    async def get(self, session_id: str, store: Store) -> bytes | None:
        """Retrieve data associated with ``session_id``.

        Args:
            session_id: The session-ID
            store: Store to retrieve the session data from

        Returns:
            The session data, if existing, otherwise ``None``.
        """
        max_age = int(self.config.max_age) if self.config.max_age is not None else None
        return await store.get(session_id, renew_for=max_age if self.config.renew_on_access else None)

    async def set(self, session_id: str, data: bytes, store: Store) -> None:
        """Store ``data`` under the ``session_id`` for later retrieval.

        If there is already data associated with ``session_id``, replace
        it with ``data`` and reset its expiry time

        Args:
            session_id: The session-ID
            data: Serialized session data
            store: Store to save the session data in

        Returns:
            None
        """
        expires_in = int(self.config.max_age) if self.config.max_age is not None else None
        await store.set(session_id, data, expires_in=expires_in)

    async def delete(self, session_id: str, store: Store) -> None:
        """Delete the data associated with ``session_id``. Fails silently if no such session-ID exists.

        Args:
            session_id: The session-ID
            store: Store to delete the session data from

        Returns:
            None
        """
        await store.delete(session_id)

    def get_session_id(self, connection: ASGIConnection) -> str:
        """Try to fetch session id from the connection. If one does not exist, generate one.

        If a session ID already exists in the cookies, it is returned.
        If there is no ID in the cookies but one in the connection state, then the session exists but has not yet
        been returned to the user.
        Otherwise, a new session must be created.

        Args:
            connection: Originating ASGIConnection containing the scope
        Returns:
            Session id str or None if the concept of a session id does not apply.
        """
        session_id = connection.cookies.get(self.config.key)
        if not session_id or session_id == "null":
            session_id = connection.get_session_id()
            if not session_id:
                session_id = self.generate_session_id()
        return session_id

    def generate_session_id(self) -> str:
        """Generate a new session-ID, with
        n=:attr:`session_id_bytes <ServerSideSessionConfig.session_id_bytes>` random bytes.

        Returns:
            A session-ID
        """
        return secrets.token_hex(self.config.session_id_bytes)

    async def store_in_message(self, scope_session: ScopeSession, message: Message, connection: ASGIConnection) -> None:
        """Store the necessary information in the outgoing ``Message`` by setting a cookie containing the session-ID.

        If the session is empty, a null-cookie will be set. Otherwise, the serialised
        data will be stored using :meth:`set <ServerSideSessionBackend.set>`, under the current session-id. If no session-ID
        exists, a new ID will be generated using :meth:`generate_session_id <ServerSideSessionBackend.generate_session_id>`.

        Args:
            scope_session: Current session to store
            message: Outgoing send-message
            connection: Originating ASGIConnection containing the scope

        Returns:
            None
        """
        scope = connection.scope
        store = self.config.get_store_from_app(scope["app"])
        headers = MutableScopeHeaders.from_message(message)
        session_id = self.get_session_id(connection)

        cookie_params = dict(extract_dataclass_items(self.config, exclude_none=True, include=Cookie.__dict__.keys()))

        if scope_session is Empty:
            await self.delete(session_id, store=store)
            headers.add(
                "Set-Cookie",
                Cookie(value="null", key=self.config.key, expires=0, **cookie_params).to_header(header=""),
            )
        else:
            serialised_data = self.serialize_data(scope_session, scope)
            await self.set(session_id=session_id, data=serialised_data, store=store)
            headers.add(
                "Set-Cookie", Cookie(value=session_id, key=self.config.key, **cookie_params).to_header(header="")
            )

    async def load_from_connection(self, connection: ASGIConnection) -> dict[str, Any]:
        """Load session data from a connection and return it as a dictionary to be used in the current application
        scope.

        The session-ID will be gathered from a cookie with the key set in
        :attr:`BaseBackendConfig.key`. If a cookie is found, its value will be used as the session-ID and data associated
        with this ID will be loaded using :meth:`get <ServerSideSessionBackend.get>`.
        If no cookie was found or no data was loaded from the store, this will return an
        empty dictionary.

        Args:
            connection: An ASGIConnection instance

        Returns:
            The current session data
        """
        if session_id := connection.cookies.get(self.config.key):
            store = self.config.get_store_from_app(connection.scope["app"])
            data = await self.get(session_id, store=store)
            if data is not None:
                return self.deserialize_data(data)
        return {}


@dataclass
class ServerSideSessionConfig(BaseBackendConfig[ServerSideSessionBackend]):  # pyright: ignore
    """Base configuration for server side backends."""

    _backend_class = ServerSideSessionBackend

    session_id_bytes: int = field(default=32)
    """Number of bytes used to generate a random session-ID."""
    renew_on_access: bool = field(default=False)
    """Renew expiry times of sessions when they're being accessed"""
    key: str = field(default="session")
    """Key to use for the cookie inside the header, e.g. ``session=<data>`` where ``session`` is the cookie key and
    ``<data>`` is the session data.

    Notes:
        - If a session cookie exceeds 4KB in size it is split. In this case the key will be of the format
          ``session-{segment number}``.

    """
    max_age: int = field(default=ONE_DAY_IN_SECONDS * 14)
    """Maximal age of the cookie before its invalidated."""
    scopes: Scopes = field(default_factory=lambda: {ScopeType.HTTP, ScopeType.WEBSOCKET})
    """Scopes for the middleware - options are ``http`` and ``websocket`` with the default being both"""
    path: str = field(default="/")
    """Path fragment that must exist in the request url for the cookie to be valid.

    Defaults to ``'/'``.
    """
    domain: str | None = field(default=None)
    """Domain for which the cookie is valid."""
    secure: bool = field(default=False)
    """Https is required for the cookie."""
    httponly: bool = field(default=True)
    """Forbids javascript to access the cookie via 'Document.cookie'."""
    samesite: Literal["lax", "strict", "none"] = field(default="lax")
    """Controls whether or not a cookie is sent with cross-site requests. Defaults to ``lax``."""
    exclude: str | list[str] | None = field(default=None)
    """A pattern or list of patterns to skip in the session middleware."""
    exclude_opt_key: str = field(default="skip_session")
    """An identifier to use on routes to disable the session middleware for a particular route."""
    store: str = "sessions"
    """Name of the :class:`Store <.stores.base.Store>` to use"""

    def __post_init__(self) -> None:
        if len(self.key) < 1 or len(self.key) > 256:
            raise ImproperlyConfiguredException("key must be a string with a length between 1-256")
        if self.max_age < 1:
            raise ImproperlyConfiguredException("max_age must be greater than 0")

    def get_store_from_app(self, app: Litestar) -> Store:
        """Get the store defined in :attr:`store` from an :class:`Litestar <.app.Litestar>` instance"""
        return app.stores.get(self.store)



from __future__ import annotations

from typing import TYPE_CHECKING, Literal

from litestar.enums import CompressionEncoding
from litestar.exceptions import MissingDependencyException
from litestar.middleware.compression.facade import CompressionFacade

try:
    from brotli import MODE_FONT, MODE_GENERIC, MODE_TEXT, Compressor
except ImportError as e:
    raise MissingDependencyException("brotli") from e


if TYPE_CHECKING:
    from io import BytesIO

    from litestar.config.compression import CompressionConfig


class BrotliCompression(CompressionFacade):
    __slots__ = ("compressor", "buffer", "compression_encoding")

    encoding = CompressionEncoding.BROTLI

    def __init__(
        self,
        buffer: BytesIO,
        compression_encoding: Literal[CompressionEncoding.BROTLI] | str,
        config: CompressionConfig,
    ) -> None:
        self.buffer = buffer
        self.compression_encoding = compression_encoding
        modes: dict[Literal["generic", "text", "font"], int] = {
            "text": int(MODE_TEXT),
            "font": int(MODE_FONT),
            "generic": int(MODE_GENERIC),
        }
        self.compressor = Compressor(
            quality=config.brotli_quality,
            mode=modes[config.brotli_mode],
            lgwin=config.brotli_lgwin,
            lgblock=config.brotli_lgblock,
        )

    def write(self, body: bytes) -> None:
        self.buffer.write(self.compressor.process(body))
        self.buffer.write(self.compressor.flush())

    def close(self) -> None:
        self.buffer.write(self.compressor.finish())



from __future__ import annotations

from copy import deepcopy
from dataclasses import dataclass, field, fields
from typing import TYPE_CHECKING, Final, Literal, Sequence

from litestar._openapi.utils import default_operation_id_creator
from litestar.openapi.plugins import (
    JsonRenderPlugin,
    RapidocRenderPlugin,
    RedocRenderPlugin,
    StoplightRenderPlugin,
    SwaggerRenderPlugin,
    YamlRenderPlugin,
)
from litestar.openapi.spec import (
    Components,
    Contact,
    ExternalDocumentation,
    Info,
    License,
    OpenAPI,
    PathItem,
    Reference,
    SecurityRequirement,
    Server,
    Tag,
)
from litestar.utils.deprecation import warn_deprecation
from litestar.utils.path import normalize_path

if TYPE_CHECKING:
    from litestar.openapi.controller import OpenAPIController
    from litestar.openapi.plugins import OpenAPIRenderPlugin
    from litestar.router import Router
    from litestar.types.callable_types import OperationIDCreator

__all__ = ("OpenAPIConfig",)

_enabled_plugin_map = {
    "elements": StoplightRenderPlugin,
    "openapi.json": JsonRenderPlugin,
    "openapi.yaml": YamlRenderPlugin,
    "openapi.yml": YamlRenderPlugin,
    "rapidoc": RapidocRenderPlugin,
    "redoc": RedocRenderPlugin,
    "swagger": SwaggerRenderPlugin,
    "oauth2-redirect.html": None,
}

_DEFAULT_SCHEMA_SITE: Final = "redoc"


@dataclass
class OpenAPIConfig:
    """Configuration for OpenAPI.

    To enable OpenAPI schema generation and serving, pass an instance of this class to the
    :class:`Litestar <.app.Litestar>` constructor using the ``openapi_config`` kwargs.
    """

    title: str
    """Title of API documentation."""
    version: str
    """API version, e.g. '1.0.0'."""

    create_examples: bool = field(default=False)
    """Generate examples using the polyfactory library."""
    random_seed: int = 10
    """The random seed used when creating the examples to ensure deterministic generation of examples."""
    contact: Contact | None = field(default=None)
    """API contact information, should be an :class:`Contact <litestar.openapi.spec.contact.Contact>` instance."""
    description: str | None = field(default=None)
    """API description."""
    external_docs: ExternalDocumentation | None = field(default=None)
    """Links to external documentation.

    Should be an instance of :class:`ExternalDocumentation <litestar.openapi.spec.external_documentation.ExternalDocumentation>`.
    """
    license: License | None = field(default=None)
    """API Licensing information.

    Should be an instance of :class:`License <litestar.openapi.spec.license.License>`.
    """
    security: list[SecurityRequirement] | None = field(default=None)
    """API Security requirements information.

    Should be an instance of
        :data:`SecurityRequirement <.openapi.spec.SecurityRequirement>`.
    """
    components: Components | list[Components] = field(default_factory=Components)
    """API Components information.

    Should be an instance of :class:`Components <litestar.openapi.spec.components.Components>` or a list thereof.
    """
    servers: list[Server] = field(default_factory=lambda: [Server(url="/")])
    """A list of :class:`Server <litestar.openapi.spec.server.Server>` instances."""
    summary: str | None = field(default=None)
    """A summary text."""
    tags: list[Tag] | None = field(default=None)
    """A list of :class:`Tag <litestar.openapi.spec.tag.Tag>` instances."""
    terms_of_service: str | None = field(default=None)
    """URL to page that contains terms of service."""
    use_handler_docstrings: bool = field(default=False)
    """Draw operation description from route handler docstring if not otherwise provided."""
    webhooks: dict[str, PathItem | Reference] | None = field(default=None)
    """A mapping of key to either :class:`PathItem <litestar.openapi.spec.path_item.PathItem>` or.

    :class:`Reference <litestar.openapi.spec.reference.Reference>` objects.
    """
    operation_id_creator: OperationIDCreator = default_operation_id_creator
    """A callable that generates unique operation ids"""
    path: str | None = field(default=None)
    """Base path for the OpenAPI documentation endpoints.

    If no path is provided the default is ``/schema``.

    Ignored if :attr:`openapi_router` is provided.
    """
    render_plugins: Sequence[OpenAPIRenderPlugin] = field(default=())
    """Plugins for rendering OpenAPI documentation UIs."""
    openapi_router: Router | None = None
    """An optional router for serving OpenAPI documentation and schema files.

    If provided, ``path`` is ignored.

    This parameter is also ignored if the deprecated :attr:`openapi_router <.openapi.OpenAPIConfig.openapi_controller>`
    kwarg is provided.

    :attr:`openapi_router` is not required, but it can be passed to customize the configuration of the router used to
    serve the documentation endpoints. For example, you can add middleware or guards to the router.

    Handlers to serve the OpenAPI schema and documentation sites are added to this router according to
    :attr:`render_plugins`, so routes shouldn't be added that conflict with these.
    """
    openapi_controller: type[OpenAPIController] | None = None
    """Controller for generating OpenAPI routes.

    Must be subclass of :class:`OpenAPIController <litestar.openapi.controller.OpenAPIController>`.

    .. deprecated:: v2.8.0
    """
    root_schema_site: Literal["redoc", "swagger", "elements", "rapidoc"] | None = None
    """The static schema generator to use for the "root" path of ``/schema/``.

    .. deprecated:: v2.8.0
    """
    enabled_endpoints: set[str] | None = None
    """A set of the enabled documentation sites and schema download endpoints.

    .. deprecated:: v2.8.0
    """

    def __post_init__(self) -> None:
        self._issue_deprecations()

        self.root_schema_site = self.root_schema_site or _DEFAULT_SCHEMA_SITE

        self.enabled_endpoints = (
            set(_enabled_plugin_map.keys()) if self.enabled_endpoints is None else self.enabled_endpoints
        )

        if self.path:
            self.path = normalize_path(self.path)

        if self.path and self.openapi_controller is not None:
            self.openapi_controller = type("OpenAPIController", (self.openapi_controller,), {"path": self.path})

        self.default_plugin: OpenAPIRenderPlugin | None = None
        if self.openapi_controller is None:
            if not self.render_plugins:
                self._plugin_backward_compatibility()
            else:
                # user is implicitly opted into the future plugin-based OpenAPI implementation
                # behavior by explicitly providing a list of render plugins
                for plugin in self.render_plugins:
                    if plugin.has_path("/"):
                        self.default_plugin = plugin
                        break
                else:
                    self.default_plugin = self.render_plugins[0]

    def _issue_deprecations(self) -> None:
        """Handle deprecated config options."""
        deprecated_in = "v2.8.0"
        removed_in = "v3.0.0"
        if self.openapi_controller is not None:
            warn_deprecation(
                deprecated_in,
                "openapi_controller",
                "attribute",
                removal_in=removed_in,
                alternative="render_plugins",
            )

        if self.root_schema_site is not None:
            warn_deprecation(
                deprecated_in,
                "root_schema_site",
                "attribute",
                removal_in=removed_in,
                alternative="render_plugins",
                info="Any 'render_plugin' with path '/' or first 'render_plugin' in list will be served at the OpenAPI root.",
            )

        if self.enabled_endpoints is not None:
            warn_deprecation(
                deprecated_in,
                "enabled_endpoints",
                "attribute",
                removal_in=removed_in,
                alternative="render_plugins",
                info="Configure a 'render_plugin' to enable an endpoint.",
            )

    def _plugin_backward_compatibility(self) -> None:
        """Backward compatibility for the plugin-based OpenAPI implementation.

        This preserves backward compatibility with the Controller-based OpenAPI implementation.

        We add a plugin for each enabled endpoint and set the default plugin to the plugin
        that has a path ending in the value of ``root_schema_site``.
        """

        def is_default_plugin(plugin_: OpenAPIRenderPlugin) -> bool:
            """Return True if the plugin is the default plugin."""
            root_schema_site = self.root_schema_site or _DEFAULT_SCHEMA_SITE
            return any(path.endswith(root_schema_site) for path in plugin_.paths)

        self.render_plugins = rps = []
        for key in self.enabled_endpoints or ():
            if plugin_type := _enabled_plugin_map[key]:
                plugin = plugin_type()
                rps.append(plugin)
                if is_default_plugin(plugin):
                    self.default_plugin = plugin

    def to_openapi_schema(self) -> OpenAPI:
        """Return an ``OpenAPI`` instance from the values stored in ``self``.

        Returns:
            An instance of :class:`OpenAPI <litestar.openapi.spec.open_api.OpenAPI>`.
        """

        if isinstance(self.components, list):
            merged_components = Components()
            for components in self.components:
                for key in (f.name for f in fields(components)):
                    if value := getattr(components, key, None):
                        merged_value_dict = getattr(merged_components, key, {}) or {}
                        merged_value_dict.update(value)
                        setattr(merged_components, key, merged_value_dict)

            self.components = merged_components

        return OpenAPI(
            external_docs=self.external_docs,
            security=self.security,
            components=deepcopy(self.components),  # deepcopy prevents mutation of the config's components
            servers=self.servers,
            tags=self.tags,
            webhooks=self.webhooks,
            info=Info(
                title=self.title,
                version=self.version,
                description=self.description,
                contact=self.contact,
                license=self.license,
                summary=self.summary,
                terms_of_service=self.terms_of_service,
            ),
            paths={},
        )



from __future__ import annotations

from typing import TYPE_CHECKING, Any, Generic, TypeVar, cast

from litestar._parsers import parse_cookie_string, parse_query_string
from litestar.datastructures.headers import Headers
from litestar.datastructures.multi_dicts import MultiDict
from litestar.datastructures.state import State
from litestar.datastructures.url import URL, Address, make_absolute_url
from litestar.exceptions import ImproperlyConfiguredException
from litestar.types.empty import Empty
from litestar.utils.empty import value_or_default
from litestar.utils.scope.state import ScopeState

if TYPE_CHECKING:
    from typing import NoReturn

    from litestar.app import Litestar
    from litestar.types import DataContainerType, EmptyType
    from litestar.types.asgi_types import Message, Receive, Scope, Send
    from litestar.types.protocols import Logger

__all__ = ("ASGIConnection", "empty_receive", "empty_send")

UserT = TypeVar("UserT")
AuthT = TypeVar("AuthT")
HandlerT = TypeVar("HandlerT")
StateT = TypeVar("StateT", bound=State)


async def empty_receive() -> NoReturn:  # pragma: no cover
    """Raise a ``RuntimeError``.

    Serves as a placeholder ``send`` function.

    Raises:
        RuntimeError
    """
    raise RuntimeError()


async def empty_send(_: Message) -> NoReturn:  # pragma: no cover
    """Raise a ``RuntimeError``.

    Serves as a placeholder ``send`` function.

    Args:
        _: An ASGI message

    Raises:
        RuntimeError
    """
    raise RuntimeError()


class ASGIConnection(Generic[HandlerT, UserT, AuthT, StateT]):
    """The base ASGI connection container."""

    __slots__ = (
        "scope",
        "receive",
        "send",
        "_base_url",
        "_url",
        "_parsed_query",
        "_cookies",
        "_server_extensions",
        "_connection_state",
    )

    scope: Scope
    """The ASGI scope attached to the connection."""
    receive: Receive
    """The ASGI receive function."""
    send: Send
    """The ASGI send function."""

    def __init__(self, scope: Scope, receive: Receive = empty_receive, send: Send = empty_send) -> None:
        """Initialize ``ASGIConnection``.

        Args:
            scope: The ASGI connection scope.
            receive: The ASGI receive function.
            send: The ASGI send function.
        """
        self.scope = scope
        self.receive = receive
        self.send = send
        self._connection_state = ScopeState.from_scope(scope)
        self._base_url: URL | EmptyType = Empty
        self._url: URL | EmptyType = Empty
        self._parsed_query: tuple[tuple[str, str], ...] | EmptyType = Empty
        self._cookies: dict[str, str] | EmptyType = Empty
        self._server_extensions = scope.get("extensions") or {}  # extensions may be None

    @property
    def app(self) -> Litestar:
        """Return the ``app`` for this connection.

        Returns:
            The :class:`Litestar <litestar.app.Litestar>` application instance
        """
        return self.scope["app"]

    @property
    def route_handler(self) -> HandlerT:
        """Return the ``route_handler`` for this connection.

        Returns:
            The target route handler instance.
        """
        return cast("HandlerT", self.scope["route_handler"])

    @property
    def state(self) -> StateT:
        """Return the ``State`` of this connection.

        Returns:
            A State instance constructed from the scope["state"] value.
        """
        return cast("StateT", State(self.scope.get("state")))

    @property
    def url(self) -> URL:
        """Return the URL of this connection's ``Scope``.

        Returns:
            A URL instance constructed from the request's scope.
        """
        if self._url is Empty:
            if (url := self._connection_state.url) is not Empty:
                self._url = url
            else:
                self._connection_state.url = self._url = URL.from_scope(self.scope)

        return self._url

    @property
    def base_url(self) -> URL:
        """Return the base URL of this connection's ``Scope``.

        Returns:
            A URL instance constructed from the request's scope, representing only the base part
            (host + domain + prefix) of the request.
        """
        if self._base_url is Empty:
            if (base_url := self._connection_state.base_url) is not Empty:
                self._base_url = base_url
            else:
                scope = cast(
                    "Scope",
                    {
                        **self.scope,
                        "path": "/",
                        "query_string": b"",
                        "root_path": self.scope.get("app_root_path") or self.scope.get("root_path", ""),
                    },
                )
                self._connection_state.base_url = self._base_url = URL.from_scope(scope)
        return self._base_url

    @property
    def headers(self) -> Headers:
        """Return the headers of this connection's ``Scope``.

        Returns:
            A Headers instance with the request's scope["headers"] value.
        """
        return Headers.from_scope(self.scope)

    @property
    def query_params(self) -> MultiDict[Any]:
        """Return the query parameters of this connection's ``Scope``.

        Returns:
            A normalized dict of query parameters. Multiple values for the same key are returned as a list.
        """
        if self._parsed_query is Empty:
            if (parsed_query := self._connection_state.parsed_query) is not Empty:
                self._parsed_query = parsed_query
            else:
                self._connection_state.parsed_query = self._parsed_query = parse_query_string(
                    self.scope.get("query_string", b"")
                )
        return MultiDict(self._parsed_query)

    @property
    def path_params(self) -> dict[str, Any]:
        """Return the ``path_params`` of this connection's ``Scope``.

        Returns:
            A string keyed dictionary of path parameter values.
        """
        return self.scope["path_params"]

    @property
    def cookies(self) -> dict[str, str]:
        """Return the ``cookies`` of this connection's ``Scope``.

        Returns:
            Returns any cookies stored in the header as a parsed dictionary.
        """
        if self._cookies is Empty:
            if (cookies := self._connection_state.cookies) is not Empty:
                self._cookies = cookies
            else:
                self._connection_state.cookies = self._cookies = (
                    parse_cookie_string(cookie_header) if (cookie_header := self.headers.get("cookie")) else {}
                )
        return self._cookies

    @property
    def client(self) -> Address | None:
        """Return the ``client`` data of this connection's ``Scope``.

        Returns:
            A two tuple of the host name and port number.
        """
        client = self.scope.get("client")
        return Address(*client) if client else None

    @property
    def auth(self) -> AuthT:
        """Return the ``auth`` data of this connection's ``Scope``.

        Raises:
            ImproperlyConfiguredException: If ``auth`` is not set in scope via an ``AuthMiddleware``, raises an exception

        Returns:
            A type correlating to the generic variable Auth.
        """
        if "auth" not in self.scope:
            raise ImproperlyConfiguredException("'auth' is not defined in scope, install an AuthMiddleware to set it")

        return cast("AuthT", self.scope["auth"])

    @property
    def user(self) -> UserT:
        """Return the ``user`` data of this connection's ``Scope``.

        Raises:
            ImproperlyConfiguredException: If ``user`` is not set in scope via an ``AuthMiddleware``, raises an exception

        Returns:
            A type correlating to the generic variable User.
        """
        if "user" not in self.scope:
            raise ImproperlyConfiguredException("'user' is not defined in scope, install an AuthMiddleware to set it")

        return cast("UserT", self.scope["user"])

    @property
    def session(self) -> dict[str, Any]:
        """Return the session for this connection if a session was previously set in the ``Scope``

        Returns:
            A dictionary representing the session value - if existing.

        Raises:
            ImproperlyConfiguredException: if session is not set in scope.
        """
        if "session" not in self.scope:
            raise ImproperlyConfiguredException(
                "'session' is not defined in scope, install a SessionMiddleware to set it"
            )

        return cast("dict[str, Any]", self.scope["session"])

    @property
    def logger(self) -> Logger:
        """Return the ``Logger`` instance for this connection.

        Returns:
            A ``Logger`` instance.

        Raises:
            ImproperlyConfiguredException: if ``log_config`` has not been passed to the Litestar constructor.
        """
        return self.app.get_logger()

    def set_session(self, value: dict[str, Any] | DataContainerType | EmptyType) -> None:
        """Set the session in the connection's ``Scope``.

        If the :class:`SessionMiddleware <.middleware.session.base.SessionMiddleware>` is enabled, the session will be added
        to the response as a cookie header.

        Args:
            value: Dictionary or pydantic model instance for the session data.

        Returns:
            None
        """
        self.scope["session"] = value

    def clear_session(self) -> None:
        """Remove the session from the connection's ``Scope``.

        If the :class:`Litestar SessionMiddleware <.middleware.session.base.SessionMiddleware>` is enabled, this will cause
        the session data to be cleared.

        Returns:
            None.
        """
        self.scope["session"] = Empty
        self._connection_state.session_id = Empty

    def get_session_id(self) -> str | None:
        return value_or_default(value=self._connection_state.session_id, default=None)

    def url_for(self, name: str, **path_parameters: Any) -> str:
        """Return the url for a given route handler name.

        Args:
            name: The ``name`` of the request route handler.
            **path_parameters: Values for path parameters in the route

        Raises:
            NoRouteMatchFoundException: If route with ``name`` does not exist, path parameters are missing or have a
                wrong type.

        Returns:
            A string representing the absolute url of the route handler.
        """
        litestar_instance = self.scope["app"]
        url_path = litestar_instance.route_reverse(name, **path_parameters)

        return make_absolute_url(url_path, self.base_url)

    def url_for_static_asset(self, name: str, file_path: str) -> str:
        """Receives a static files handler name, an asset file path and returns resolved absolute url to the asset.

        Args:
            name: A static handler unique name.
            file_path: a string containing path to an asset.

        Raises:
            NoRouteMatchFoundException: If static files handler with ``name`` does not exist.

        Returns:
            A string representing absolute url to the asset.
        """
        litestar_instance = self.scope["app"]
        url_path = litestar_instance.url_for_static_asset(name, file_path)

        return make_absolute_url(url_path, self.base_url)



import logging
from dataclasses import dataclass
from typing import TYPE_CHECKING, Any, Awaitable, Callable, List, cast

import pytest
from starlette.middleware.base import BaseHTTPMiddleware, RequestResponseEndpoint

from litestar import Controller, Request, Response, Router, get, post
from litestar.enums import ScopeType
from litestar.middleware import DefineMiddleware, MiddlewareProtocol
from litestar.testing import create_test_client

if TYPE_CHECKING:
    from typing import Type

    from _pytest.logging import LogCaptureFixture

    from litestar.types import ASGIApp, Receive, Scope, Send

logger = logging.getLogger(__name__)


class MiddlewareProtocolRequestLoggingMiddleware(MiddlewareProtocol):
    def __init__(self, app: "ASGIApp", kwarg: str = "") -> None:
        self.app = app
        self.kwarg = kwarg

    async def __call__(self, scope: "Scope", receive: "Receive", send: "Send") -> None:
        if scope["type"] == ScopeType.HTTP:
            request = Request[Any, Any, Any](scope=scope, receive=receive)
            body = await request.json()
            logger.info(f"test logging: {request.method}, {request.url}, {body}")
        await self.app(scope, receive, send)


class BaseMiddlewareRequestLoggingMiddleware(BaseHTTPMiddleware):
    async def dispatch(self, request: Request, call_next: RequestResponseEndpoint) -> Response:  # type: ignore[explicit-override, override]
        logging.getLogger(__name__).info("%s - %s", request.method, request.url)
        return await call_next(request)  # type: ignore[arg-type, return-value]


class MiddlewareWithArgsAndKwargs(BaseHTTPMiddleware):
    def __init__(self, arg: int = 0, *, app: Any, kwarg: str) -> None:
        super().__init__(app)
        self.arg = arg
        self.kwarg = kwarg

    async def dispatch(  # type: ignore[empty-body, explicit-override, override]
        self, request: Request, call_next: Callable[[Request], Awaitable[Response]]
    ) -> Response: ...


@pytest.mark.parametrize(
    "middleware",
    [
        BaseMiddlewareRequestLoggingMiddleware,
        # Middleware(MiddlewareWithArgsAndKwargs, kwarg="123Jeronimo"),  # pyright: ignore[reportGeneralTypeIssues] # noqa: ERA001
        # Middleware(MiddlewareProtocolRequestLoggingMiddleware, kwarg="123Jeronimo"),  # type: ignore[arg-type] # pyright: ignore[reportGeneralTypeIssues] # noqa: ERA001
        DefineMiddleware(MiddlewareWithArgsAndKwargs, 1, kwarg="123Jeronimo"),  # type: ignore[arg-type]
        DefineMiddleware(MiddlewareProtocolRequestLoggingMiddleware, kwarg="123Jeronimo"),
    ],
)
def test_custom_middleware_processing(middleware: Any) -> None:
    @get(path="/")
    def handler() -> None: ...

    with create_test_client(route_handlers=[handler], middleware=[middleware]) as client:
        app = client.app
        assert app.middleware == [middleware]

        unpacked_middleware = []
        cur = client.app.asgi_router.root_route_map_node.children["/"].asgi_handlers["GET"][0]
        while hasattr(cur, "app"):
            unpacked_middleware.append(cur)
            cur = cast("ASGIApp", cur.app)  # pyright: ignore
        unpacked_middleware.append(cur)
        assert len(unpacked_middleware) == 4

        middleware_instance = unpacked_middleware[1]

        assert isinstance(
            middleware_instance,
            (
                MiddlewareProtocolRequestLoggingMiddleware,
                BaseMiddlewareRequestLoggingMiddleware,
                MiddlewareWithArgsAndKwargs,
            ),
        )
        if isinstance(middleware_instance, (MiddlewareProtocolRequestLoggingMiddleware, MiddlewareWithArgsAndKwargs)):
            assert middleware_instance.kwarg == "123Jeronimo"
        if isinstance(middleware, DefineMiddleware) and isinstance(middleware_instance, MiddlewareWithArgsAndKwargs):
            assert middleware_instance.arg == 1


def test_request_body_logging_middleware(caplog: "LogCaptureFixture") -> None:
    @dataclass
    class JSONRequest:
        name: str
        age: int
        programmer: bool

    @post(path="/")
    def post_handler(data: JSONRequest) -> JSONRequest:
        return data

    with caplog.at_level(logging.INFO):
        client = create_test_client(
            route_handlers=[post_handler], middleware=[MiddlewareProtocolRequestLoggingMiddleware]
        )
        response = client.post("/", json={"name": "moishe zuchmir", "age": 40, "programmer": True})
        assert response.status_code == 201
        assert "test logging" in caplog.text


def test_middleware_call_order() -> None:
    """Test that middlewares are called in the order they have been passed."""

    results: List[int] = []

    def create_test_middleware(middleware_id: int) -> "Type[MiddlewareProtocol]":
        class TestMiddleware(MiddlewareProtocol):
            def __init__(self, app: "ASGIApp") -> None:
                self.app = app

            async def __call__(self, scope: "Scope", receive: "Receive", send: "Send") -> None:
                results.append(middleware_id)
                await self.app(scope, receive, send)

        return TestMiddleware

    class MyController(Controller):
        path = "/controller"
        middleware = [create_test_middleware(4), create_test_middleware(5)]

        @get("/handler", middleware=[create_test_middleware(6), create_test_middleware(7)])
        def my_handler(self) -> None:
            return None

    router = Router(
        path="/router", route_handlers=[MyController], middleware=[create_test_middleware(2), create_test_middleware(3)]
    )

    with create_test_client(
        route_handlers=[router],
        middleware=[create_test_middleware(0), create_test_middleware(1)],
    ) as client:
        client.get("/router/controller/handler")

        assert results == [0, 1, 2, 3, 4, 5, 6, 7]



from __future__ import annotations

from dataclasses import replace
from typing import TYPE_CHECKING, Any, Collection, Generic, TypeVar
from warnings import warn

from typing_extensions import Annotated, TypeAlias, override

from litestar.contrib.pydantic.utils import is_pydantic_undefined
from litestar.dto.base_dto import AbstractDTO
from litestar.dto.data_structures import DTOFieldDefinition
from litestar.dto.field import DTO_FIELD_META_KEY, extract_dto_field
from litestar.exceptions import MissingDependencyException, ValidationException
from litestar.types.empty import Empty
from litestar.typing import FieldDefinition

if TYPE_CHECKING:
    from typing import Generator

try:
    import pydantic as _  # noqa: F401
except ImportError as e:
    raise MissingDependencyException("pydantic") from e


try:
    import pydantic as pydantic_v2

    assert pydantic_v2.__version__.startswith("2.")  # noqa: S101

    from pydantic import ValidationError as ValidationErrorV2
    from pydantic import v1 as pydantic_v1
    from pydantic.v1 import ValidationError as ValidationErrorV1

    ModelType: TypeAlias = "pydantic_v1.BaseModel | pydantic_v2.BaseModel"

except AssertionError:
    import pydantic as pydantic_v1  # type: ignore[no-redef]

    pydantic_v2 = Empty  # type: ignore[assignment]
    from pydantic import ValidationError as ValidationErrorV1  # type: ignore[assignment]

    ValidationErrorV2 = ValidationErrorV1  # type: ignore[assignment, misc]
    ModelType = "pydantic_v1.BaseModel"  # type: ignore[misc]


T = TypeVar("T", bound="ModelType | Collection[ModelType]")


__all__ = ("PydanticDTO",)

_down_types: dict[Any, Any] = {
    pydantic_v1.EmailStr: str,
    pydantic_v1.IPvAnyAddress: str,
    pydantic_v1.IPvAnyInterface: str,
    pydantic_v1.IPvAnyNetwork: str,
}

if pydantic_v2 is not Empty:  # type: ignore[comparison-overlap]  # pragma: no cover
    _down_types.update(
        {
            pydantic_v2.JsonValue: Any,
            pydantic_v2.EmailStr: str,
            pydantic_v2.IPvAnyAddress: str,
            pydantic_v2.IPvAnyInterface: str,
            pydantic_v2.IPvAnyNetwork: str,
        }
    )


def convert_validation_error(validation_error: ValidationErrorV1 | ValidationErrorV2) -> list[dict[str, Any]]:
    error_list = validation_error.errors()
    for error in error_list:
        if isinstance(exception := error.get("ctx", {}).get("error"), Exception):
            error["ctx"]["error"] = type(exception).__name__
    return error_list  # type: ignore[return-value]


def downtype_for_data_transfer(field_definition: FieldDefinition) -> FieldDefinition:
    if sub := _down_types.get(field_definition.annotation):
        return FieldDefinition.from_kwarg(
            annotation=Annotated[sub, field_definition.metadata], name=field_definition.name
        )
    return field_definition


class PydanticDTO(AbstractDTO[T], Generic[T]):
    """Support for domain modelling with Pydantic."""

    @override
    def decode_builtins(self, value: dict[str, Any]) -> Any:
        try:
            return super().decode_builtins(value)
        except (ValidationErrorV2, ValidationErrorV1) as ex:
            raise ValidationException(extra=convert_validation_error(ex)) from ex

    @override
    def decode_bytes(self, value: bytes) -> Any:
        try:
            return super().decode_bytes(value)
        except (ValidationErrorV2, ValidationErrorV1) as ex:
            raise ValidationException(extra=convert_validation_error(ex)) from ex

    @classmethod
    def generate_field_definitions(
        cls, model_type: type[pydantic_v1.BaseModel | pydantic_v2.BaseModel]
    ) -> Generator[DTOFieldDefinition, None, None]:
        model_field_definitions = cls.get_model_type_hints(model_type)

        model_fields: dict[str, pydantic_v1.fields.FieldInfo | pydantic_v2.fields.FieldInfo]
        try:
            model_fields = dict(model_type.model_fields)  # type: ignore[union-attr]
        except AttributeError:
            model_fields = {
                k: model_field.field_info
                for k, model_field in model_type.__fields__.items()  # type: ignore[union-attr]
            }

        for field_name, field_info in model_fields.items():
            field_definition = downtype_for_data_transfer(model_field_definitions[field_name])
            dto_field = extract_dto_field(field_definition, field_definition.extra)

            try:
                extra = field_info.extra  # type: ignore[union-attr]
            except AttributeError:
                extra = field_info.json_schema_extra  # type: ignore[union-attr]

            if extra is not None and extra.pop(DTO_FIELD_META_KEY, None):
                warn(
                    message="Declaring 'DTOField' via Pydantic's 'Field.extra' is deprecated. "
                    "Use 'Annotated', e.g., 'Annotated[str, DTOField(mark='read-only')]' instead. "
                    "Support for 'DTOField' in 'Field.extra' will be removed in v3.",
                    category=DeprecationWarning,
                    stacklevel=2,
                )

            if not is_pydantic_undefined(field_info.default):
                default = field_info.default
            elif field_definition.is_optional:
                default = None
            else:
                default = Empty

            yield replace(
                DTOFieldDefinition.from_field_definition(
                    field_definition=field_definition,
                    dto_field=dto_field,
                    model_name=model_type.__name__,
                    default_factory=field_info.default_factory
                    if field_info.default_factory and not is_pydantic_undefined(field_info.default_factory)
                    else None,
                ),
                default=default,
                name=field_name,
            )

    @classmethod
    def detect_nested_field(cls, field_definition: FieldDefinition) -> bool:
        if pydantic_v2 is not Empty:  # type: ignore[comparison-overlap]
            return field_definition.is_subclass_of((pydantic_v1.BaseModel, pydantic_v2.BaseModel))
        return field_definition.is_subclass_of(pydantic_v1.BaseModel)  # type: ignore[unreachable]



from datetime import time
from typing import Type

import pytest

from litestar import Litestar, Router, delete, get, patch, post, put
from litestar.exceptions import NoRouteMatchFoundException
from litestar.handlers.http_handlers import HTTPRouteHandler


@pytest.mark.parametrize("decorator", [get, post, patch, put, delete])
def test_route_reverse(decorator: Type[HTTPRouteHandler]) -> None:
    @decorator("/path-one/{param:str}", name="handler-name")  # type: ignore[call-arg]
    def handler() -> None:
        return None

    @decorator("/path-two", name="handler-no-params")  # type: ignore[call-arg]
    def handler_no_params() -> None:
        return None

    @decorator("/multiple/{str_param:str}/params/{int_param:int}/", name="multiple-params-handler-name")  # type: ignore[call-arg]
    def handler2() -> None:
        return None

    @decorator(
        ["/handler3", "/handler3/{str_param:str}/", "/handler3/{str_param:str}/{int_param:int}/"],
        name="multiple-default-params",
    )  # type: ignore[call-arg]
    def handler3(str_param: str = "default", int_param: int = 0) -> None:
        return None

    @decorator(["/handler4/int/{int_param:int}", "/handler4/str/{str_param:str}"], name="handler4")  # type: ignore[call-arg]
    def handler4(int_param: int = 1, str_param: str = "str") -> None:
        return None

    router = Router("router-path/", route_handlers=[handler, handler_no_params, handler3, handler4])
    router_with_param = Router("router-with-param/{router_param:str}", route_handlers=[handler2])
    app = Litestar(route_handlers=[router, router_with_param])

    reversed_url_path = app.route_reverse("handler-name", param="param-value")
    assert reversed_url_path == "/router-path/path-one/param-value"

    reversed_url_path = app.route_reverse("handler-no-params")
    assert reversed_url_path == "/router-path/path-two"

    reversed_url_path = app.route_reverse(
        "multiple-params-handler-name", router_param="router", str_param="abc", int_param=123
    )
    assert reversed_url_path == "/router-with-param/router/multiple/abc/params/123"

    reversed_url_path = app.route_reverse("handler4", int_param=100)
    assert reversed_url_path == "/router-path/handler4/int/100"

    reversed_url_path = app.route_reverse("handler4", str_param="string")
    assert reversed_url_path == "/router-path/handler4/str/string"

    with pytest.raises(NoRouteMatchFoundException):
        reversed_url_path = app.route_reverse("nonexistent-handler")


@pytest.mark.parametrize(
    "complex_path_param",
    [("time", time(hour=14), "14:00"), ("float", float(1 / 3), "0.33")],
)
def test_route_reverse_validation_complex_params(complex_path_param) -> None:  # type: ignore[no-untyped-def]
    param_type, param_value, param_manual_str = complex_path_param

    @get(f"/abc/{{param:{param_type}}}", name="handler")
    def handler() -> None:
        pass

    app = Litestar(route_handlers=[handler])

    # test that complex types of path params accept either itself
    # or string but nothing else
    with pytest.raises(NoRouteMatchFoundException):
        app.route_reverse("handler", param=123)

    reversed_url_path = app.route_reverse("handler", param=param_manual_str)
    assert reversed_url_path == f"/abc/{param_manual_str}"

    reversed_url_path = app.route_reverse("handler", param=param_value)
    assert reversed_url_path == f"/abc/{param_value}"


def test_route_reverse_validation() -> None:
    @get("/abc/{param:int}", name="handler-name")
    def handler_one() -> None:
        pass

    @get("/def/{param:str}", name="another-handler-name")
    def handler_two() -> None:
        pass

    app = Litestar(route_handlers=[handler_one, handler_two])

    with pytest.raises(NoRouteMatchFoundException):
        app.route_reverse("handler-name")

    with pytest.raises(NoRouteMatchFoundException):
        app.route_reverse("handler-name", param="str")

    with pytest.raises(NoRouteMatchFoundException):
        app.route_reverse("another-handler-name", param=1)



from litestar.utils.sync import ensure_async_callable


async def test_function_wrapper_wraps_method_correctly() -> None:
    class MyClass:
        def __init__(self) -> None:
            self.value = 0

        def my_method(self, value: int) -> None:
            self.value = value

    instance = MyClass()

    wrapped_method = ensure_async_callable(instance.my_method)

    await wrapped_method(1)
    assert instance.value == 1

    await wrapped_method(value=10)
    assert instance.value == 10


async def test_function_wrapper_wraps_async_method_correctly() -> None:
    class MyClass:
        def __init__(self) -> None:
            self.value = 0

        async def my_method(self, value: int) -> None:
            self.value = value

    instance = MyClass()

    wrapped_method = ensure_async_callable(instance.my_method)

    await wrapped_method(1)  # type: ignore[unused-coroutine]
    assert instance.value == 1

    await wrapped_method(value=10)  # type: ignore[unused-coroutine]
    assert instance.value == 10


async def test_function_wrapper_wraps_function_correctly() -> None:
    obj = {"value": 0}

    def my_function(new_value: int) -> None:
        obj["value"] = new_value

    wrapped_function = ensure_async_callable(my_function)

    await wrapped_function(1)
    assert obj["value"] == 1

    await wrapped_function(new_value=10)
    assert obj["value"] == 10


async def test_function_wrapper_wraps_async_function_correctly() -> None:
    obj = {"value": 0}

    async def my_function(new_value: int) -> None:
        obj["value"] = new_value

    wrapped_function = ensure_async_callable(my_function)

    await wrapped_function(1)  # type: ignore[unused-coroutine]
    assert obj["value"] == 1

    await wrapped_function(new_value=10)  # type: ignore[unused-coroutine]
    assert obj["value"] == 10


async def test_function_wrapper_wraps_class_correctly() -> None:
    class MyCallable:
        value = 0

        def __call__(self, new_value: int) -> None:
            self.value = new_value

    instance = MyCallable()

    wrapped_class = ensure_async_callable(instance)

    await wrapped_class(1)
    assert instance.value == 1

    await wrapped_class(new_value=10)
    assert instance.value == 10


async def test_function_wrapper_wraps_async_class_correctly() -> None:
    class MyCallable:
        value = 0

        async def __call__(self, new_value: int) -> None:
            self.value = new_value

    instance = MyCallable()

    wrapped_class = ensure_async_callable(instance)

    await wrapped_class(1)  # type: ignore[unused-coroutine]
    assert instance.value == 1

    await wrapped_class(new_value=10)  # type: ignore[unused-coroutine]
    assert instance.value == 10



from __future__ import annotations

import asyncio
from concurrent.futures import ThreadPoolExecutor
from typing import Generator
from unittest.mock import AsyncMock

import pytest
import trio
from pytest_mock import MockerFixture

from litestar.concurrency import (
    _State,
    get_asyncio_executor,
    get_trio_capacity_limiter,
    set_asyncio_executor,
    set_trio_capacity_limiter,
    sync_to_thread,
)


@pytest.fixture(autouse=True)
def reset_state() -> Generator[None, None, None]:
    _State.LIMITER = None
    _State.EXECUTOR = None
    yield
    _State.LIMITER = None
    _State.EXECUTOR = None


def func() -> int:
    return 1


def test_sync_to_thread_asyncio() -> None:
    loop = asyncio.new_event_loop()
    assert loop.run_until_complete(sync_to_thread(func)) == 1


def test_sync_to_thread_trio() -> None:
    assert trio.run(sync_to_thread, func) == 1


def test_get_set_asyncio_executor() -> None:
    assert get_asyncio_executor() is None
    executor = ThreadPoolExecutor()
    set_asyncio_executor(executor)
    assert get_asyncio_executor() is executor


def test_get_set_trio_capacity_limiter() -> None:
    limiter = trio.CapacityLimiter(10)
    assert get_trio_capacity_limiter() is None
    set_trio_capacity_limiter(limiter)
    assert get_trio_capacity_limiter() is limiter


def test_asyncio_uses_executor(mocker: MockerFixture) -> None:
    executor = ThreadPoolExecutor()

    mocker.patch("litestar.concurrency.get_asyncio_executor", return_value=executor)
    mock_run_in_executor = AsyncMock()
    mocker.patch("litestar.concurrency.asyncio.get_running_loop").return_value.run_in_executor = mock_run_in_executor

    loop = asyncio.new_event_loop()
    loop.run_until_complete(sync_to_thread(func))

    assert mock_run_in_executor.call_args_list[0].args[0] is executor


def test_set_asyncio_executor_from_running_loop_raises() -> None:
    async def main() -> None:
        set_asyncio_executor(ThreadPoolExecutor())

    with pytest.raises(RuntimeError):
        asyncio.new_event_loop().run_until_complete(main())

    assert get_asyncio_executor() is None


def test_trio_uses_limiter(mocker: MockerFixture) -> None:
    limiter = trio.CapacityLimiter(10)
    mocker.patch("litestar.concurrency.get_trio_capacity_limiter", return_value=limiter)
    mock_run_sync = mocker.patch("trio.to_thread.run_sync", new_callable=AsyncMock)

    trio.run(sync_to_thread, func)

    assert mock_run_sync.call_args_list[0].kwargs["limiter"] is limiter


def test_set_trio_capacity_limiter_from_async_context_raises() -> None:
    async def main() -> None:
        set_trio_capacity_limiter(trio.CapacityLimiter(1))

    with pytest.raises(RuntimeError):
        trio.run(main)

    assert get_trio_capacity_limiter() is None


def test_sync_to_thread_unsupported_lib(mocker: MockerFixture) -> None:
    mocker.patch("litestar.concurrency.sniffio.current_async_library", return_value="something")

    with pytest.raises(RuntimeError):
        asyncio.new_event_loop().run_until_complete(sync_to_thread(func))
